# -*- coding: UTF-8 -*-
# File: pascal_voc.py
# Author: jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals

import os, sys
import numpy as np
import tarfile
import time
import xml.etree.ElementTree as ET
from antgo.utils.fs import maybe_here_match_format
from antgo.dataflow.dataset.dataset import *
from antgo.framework.helper.fileio.file_client import *


__all__ = ['Pascal2007', 'Pascal2012']
class PascalBase(Dataset):
  def __init__(self, year, image_set, dir=None, ext_params=None):
    super(PascalBase, self).__init__(image_set, dir, ext_params)
    if not os.path.exists(self.dir):
      os.makedirs(self.dir)

    self._year = year
    self._image_set = image_set
    self._devkit_path = dir

    if self._year == '2007':
      if os.environ.get('LOCAL_RANK', 0) == 0:
        if not os.path.exists(os.path.join(self._devkit_path, 'VOCdevkit')):
          # 数据集不存在，需要重新下载，并创建标记
          ali = AliBackend()
          ali.download('ali:///dataset/voc/VOCtrainval_06-Nov-2007.tar', self.dir)
          ali.download('ali:///dataset/voc/VOCtest_06-Nov-2007.tar', self.dir)

          maybe_data_path = maybe_here_match_format(self._devkit_path, 'VOC' + self._year)
          if maybe_data_path is None:
            # auto untar
            tar = tarfile.open(os.path.join(self.dir,'VOCtrainval_06-Nov-2007.tar'), 'r')
            tar.extractall(self.dir)
            tar.close()

            tar = tarfile.open(os.path.join(self.dir, 'VOCtest_06-Nov-2007.tar'), 'r')
            tar.extractall(self.dir)
            tar.close()

          os.system('touch DATASET_IS_READY')
        else:
          # 数据集存在，创建标记
          if not os.path.exists('DATASET_IS_READY'):
            os.system('touch DATASET_IS_READY')
      else:
        while True:
          # 等待直到存在指定文件
          if os.path.exists('DATASET_IS_READY'):
            break
          time.sleep(5)
    else:
      if os.environ.get('LOCAL_RANK', 0) == 0:
        if not os.path.exists(os.path.join(self._devkit_path, 'VOCdevkit')):
          # 数据集不存在，需要重新下载，并创建标记
          ali = AliBackend()
          ali.download('ali:///dataset/voc/VOCtrainval_11-May-2012.tar', self.dir)
          ali.download('ali:///dataset/voc/VOC2012test.tar', self.dir)
          ali.download('ali:///dataset/voc/SegmentationClassAug.zip', self.dir)
          ali.download('ali:///dataset/voc/trainaug.txt', self.dir)
          maybe_data_path = maybe_here_match_format(self._devkit_path, 'VOC' + self._year)
          if maybe_data_path is None:
            # auto untar
            try:
              tar = tarfile.open(os.path.join(self.dir, 'VOCtrainval_11-May-2012.tar'), 'r')
              tar.extractall(self.dir)
              tar.close()
            except:
              print(f'Untar {os.path.join(self.dir, "VOCtrainval_11-May-2012.tar")} fail')

            try:
              tar = tarfile.open(os.path.join(self.dir, 'VOC2012test.tar'), 'r')
              tar.extractall(self.dir)
              tar.close()
            except:
              print(f'Untar {os.path.join(self.dir, "VOC2012test.tar")} fail')

            # unzip aug class 
            os.system(f"cd {self.dir} && unzip SegmentationClassAug.zip")
          os.system('touch DATASET_IS_READY')
        else:
          # 数据集存在，创建标记
          if not os.path.exists('DATASET_IS_READY'):
            os.system('touch DATASET_IS_READY')
      else:
          while True:
            # 等待直到存在指定文件
            if os.path.exists('DATASET_IS_READY'):
              break
            time.sleep(5)

    self._data_path = os.path.join(self.dir,'VOCdevkit', 'VOC' + self._year)
    self._classes = ('background',
                     'aeroplane',
                     'bicycle',
                     'bird',
                     'boat',
                     'bottle',
                     'bus',
                     'car',
                     'cat',
                     'chair',
                     'cow',
                     'diningtable',
                     'dog',
                     'horse',
                     'motorbike',
                     'person',
                     'pottedplant',
                     'sheep',
                     'sofa',
                     'train',
                     'tvmonitor')

    self._c_2_rgb = {0: (0, 0, 0),
                     1: (128, 0, 0),
                     2: (0, 128, 0),
                     3: (128, 128, 0),
                     4: (0, 0, 128),
                     5: (128, 0, 128),
                     6: (0, 128, 128),
                     7: (128, 128, 128),
                     8: (64, 0, 0),
                     9: (192, 0, 0),
                     10: (64, 128, 0),
                     11: (192, 128, 0),
                     12: (64, 0, 128),
                     13: (192, 0, 128),
                     14: (64, 128, 128),
                     15: (192, 128, 128),
                     16: (0, 64, 0),
                     17: (128, 64, 0),
                     18: (0, 192, 0),
                     19: (128, 192, 0),
                     20: (0, 64, 128)}
    self._c_2_rgb_index = {}
    for k, v in self._c_2_rgb.items():
      self._c_2_rgb_index[k] = v[2]*255*255+v[1]*255+v[0]

    self._action_classes = {'phoning': 1,
                            'playinginstrument': 2,
                            'reading': 3,
                            'ridingbike': 4,
                            'ridinghorse': 5,
                            'running': 6,
                            'takingphoto': 7,
                            'usingcomputer': 8,
                            'walking': 9,
                            'jumping': 10,
                            'other': 0}

    self._num_classes = len(self._classes)
    self._class_to_ind = dict(zip(self._classes, range(self._num_classes)))
    self._image_ext = '.jpg'
    self._image_index = self._load_image_set_index()
    self._image_index_num = len(self._image_index)

    # PASCAL specific config options
    self.config = {'cleanup': True,
                   'use_salt': True,
                   'use_diff': False,
                   'matlab_eval': False,
                   'rpn_file': None,
                   'min_size': 2}

    assert os.path.exists(self._data_path), 'path does not exist: {}'.format(self._data_path)

    self._aug_dataset = []
    self._aug_seg = []
    if self._year == '2012' and self.train_or_test == 'train':
      with open(os.path.join(self.dir, 'trainaug.txt'), 'r') as fp:
        content = fp.readline()

        while(content):
          content = content.strip()
          if content == "":
            break

          self._aug_dataset.append(os.path.join(self.dir, 'VOCdevkit/VOC2012/JPEGImages', '%s.jpg'%content))
          self._aug_seg.append(os.path.join(self.dir, 'SegmentationClassAug', '%s.png'%content))
          content = fp.readline()

  @property
  def size(self):
    return self._image_index_num + len(self._aug_dataset)

  def data_pool(self):
    raise NotImplemented

  def at(self, id):
    if id >= self._image_index_num:
        image_path = self._aug_dataset[id - self._image_index_num]
        image_seg_path = self._aug_seg[id - self._image_index_num]
        image = cv2.imread(image_path)
        seg_img = cv2.imread(image_seg_path, cv2.IMREAD_GRAYSCALE)
        return (
          image, 
          {
            'segments': seg_img,
            'image_meta': {
              'image_shape': (image.shape[0], image.shape[1])
            }            
          }
        )

    index = self._image_index[id]
    gt_roidb = self._load_roidb(index)
    image = cv2.imread(self.image_path_from_index(index))
    annos = {
      'bboxes': gt_roidb['bbox'].astype(np.float32),
      'labels': gt_roidb['category_id'],
      'segments': gt_roidb['segments'],
      'image_meta': {
        'image_shape': (image.shape[0], image.shape[1])
      }
    }
    return (image, annos)

  def image_path_from_index(self, index):
    """
    Construct an image path from the image's "index" identifier.
    """
    image_path = os.path.join(self._data_path, 'JPEGImages',
                              index + self._image_ext)
    assert os.path.exists(image_path), \
            'Path does not exist: {}'.format(image_path)
    return image_path

  def image_path_at(self, i):
    """
    Return the absolute path to image i in the image sequence.
    """
    return self.image_path_from_index(self._image_index[i])

  def _load_roidb(self, index):
    return self._load_pascal_annotation(index)

  def _load_image_set_index(self):
    """
    Load the indexes listed in this dataset's image set file.
    """
    # Example path to image set file:
    # self._devkit_path + /VOCdevkit2007/VOC2007/ImageSets/Main/val.txt
    image_set_file = os.path.join(self._data_path, 'ImageSets', 'Main',
                                  self._image_set + '.txt')

    image_set_file = \
      os.path.join(self._data_path, 'ImageSets', 'Segmentation', self._image_set + '.txt')
    assert os.path.exists(image_set_file), \
            'Path Does not Exist: {}'.format(image_set_file)
    with open(image_set_file) as f:
        image_index = [x.strip() for x in f.readlines()]
    return image_index

  def _load_pascal_annotation(self, index):
    """
    Load image and bounding boxes info from XML file in the PASCAL VOC
    format.
    """
    filename = os.path.join(self._data_path, 'Annotations', index + '.xml')
    tree = ET.parse(filename)
    objs = tree.findall('object')

    # Exclude the samples labeled as difficult
    # non_diff_objs = [
    #     obj for obj in objs if int(obj.find('difficult').text) == 0]
    # if len(non_diff_objs) != len(objs):
    #     print 'Removed {} difficult objects'.format(
    #         len(objs) - len(non_diff_objs))
    # objs = non_diff_objs

    num_objs = len(objs)

    boxes = np.zeros((num_objs, 4), dtype=np.uint16)
    category_id = np.zeros((num_objs), dtype=np.int32)
    # "Seg" area for pascal is just the box area
    area = np.zeros((num_objs), dtype=np.float32)
    category = []
    difficult = []

    segmented = tree.find('segmented')
    has_seg = 0
    if segmented is not None:
      has_seg = int(segmented.text)

    seg_img_index = None
    if has_seg:
      seg_file = os.path.join(self._data_path, 'SegmentationClass', index + '.png')
      seg_img = imread(seg_file)
      seg_img_int = seg_img.astype(np.int32)
      seg_img_index = seg_img_int[:,:,0]*255*255 + seg_img_int[:,:,1]*255 + seg_img_int[:,:,2]

    # Load object bb and segmentation into a data frame.
    # 目标框
    person_action = []
    for ix, obj in enumerate(objs):
      bbox = obj.find('bndbox')
      # Make pixel indexes 0-based
      x1 = float(bbox.find('xmin').text) - 1
      y1 = float(bbox.find('ymin').text) - 1
      x2 = float(bbox.find('xmax').text) - 1
      y2 = float(bbox.find('ymax').text) - 1
      cls = self._class_to_ind[obj.find('name').text.lower().strip()]
      boxes[ix, :] = [x1, y1, x2, y2]
      category_id[ix] = cls
      area[ix] = (x2 - x1 + 1) * (y2 - y1 + 1)
      category.append(obj.find('name').text.lower().strip())
      difficult.append(int(obj.find('difficult').text))

      if category[-1] == 'person':
        action_cls = obj.find('actions')
        if action_cls is not None:
          action_label = np.zeros((11), np.int32)
          for action_cc in action_cls.getchildren():
            action_label[self._action_classes[action_cc.tag]] = int(action_cc.text)

          person_action.append(action_label)
        else:
          person_action.append(None)
      else:
        person_action.append(None)

    annotation = {'bbox': boxes,
                  'category_id': category_id,
                  'category': category,
                  'flipped': False,
                  'difficult': difficult,
                  'area': area,
                  'person_action': person_action}

    if has_seg:
      # 语义分割
      segmentation = np.zeros((seg_img_index.shape[0], seg_img_index.shape[1]), np.uint8)
      for cls in set(category_id.tolist()):
        segmentation[np.where(seg_img_index == self._c_2_rgb_index[cls])] = cls
      annotation.update({'segments': segmentation})

    return annotation


class Pascal2007(PascalBase):
  def __init__(self, train_or_test, dir=None, ext_params=None):
    super(Pascal2007,self).__init__('2007', train_or_test, dir, ext_params)

  def split(self, split_params={}, split_method='holdout'):
    assert (self.train_or_test == 'train')
    assert (split_method == 'holdout')
    validation_pascal2007 = Pascal2007('val', self.dir)

    return self, validation_pascal2007


class Pascal2012(PascalBase):
  def __init__(self, train_or_test, dir=None, ext_params=None):
    super(Pascal2012,self).__init__('2012', train_or_test, dir, ext_params)

  def split(self, split_params={}, split_method='holdout'):
    assert (self.train_or_test == 'train')
    assert (split_method == 'holdout')
    validation_pascal2012 = Pascal2012('val', self.dir)

    return self, validation_pascal2012

# p2012 = Pascal2012('val', '/root/workspace/dataset/temp_dataset')
# rfi = ColorDistort()
# print(f'p2012 size {p2012.size}')
# for i in range(p2012.size):
#   result = p2012.sample(i)
#   result = rfi(result)
#   segments  = result['segments']
#   image = result['image']
#   canvas = np.zeros((image.shape[0], image.shape[1]+image.shape[1]), dtype=np.uint8)
#   canvas[:,:image.shape[1]] = image[:,:,0]
#   canvas[:,image.shape[1]:] = (segments/20*255).astype(np.uint8)
#   cv2.imwrite('./1234.png', canvas.astype(np.uint8))
#   print(i)
# value = p2012.sample(0)
# print(value.keys())
# value = p2012.sample(1)
# print(value)


# (1) 训练过程
# 多机多卡训练
# TODO
# 单机多卡训练(4卡运行)
# bash launch.sh ./cifar10/main.py 4 --exp=xxx --no-validate --process=train
# 单机1卡训练(可以自定义使用第几块卡 --gpu-id=0)
# python3 ./cifar10/main.py --exp=xxx --gpu-id=0 --no-validate --process=train
# 单机CPU训练(仅用于调试)
# python3 ./cifar10/main.py --exp=xxx --gpu-id=-1 --no-validate --process=train

# (2) 评估过程
# 单机多卡评估(4卡运行)
# bash launch.sh ./cifar10/main.py 4 --exp=xxx --checkpoint=yyy --process=test
# 单机1卡评估(可以自定义使用第几块卡 --gpu-id=0)
# python3 ./cifar10/main.py --exp=xxx --checkpoint=yyy --gpu-id=0 --process=test
# 单机CPU测试(仅用于调试)
# python3 ./cifar10/main.py --exp=xxx --checkpoint=yyy --gpu-id=-1 --process=test

# (3) 模型导出过程
# python3 ./cifar10/main.py --exp=xxx --checkpoint=yyy --process=export

# 1.step 通用模块
import shutil
import sys
import os
import json
import logging

system_path = os.path.join(os.path.abspath(os.curdir),'system.py')
os.system(f'ln -sf {system_path}  {os.path.dirname(os.path.realpath(__file__))}/system.py')
import torch
from antgo.utils import args
from antgo.framework.helper.trainer import *
from antgo.framework.helper.tester import *
from antgo.framework.helper.activelearning import Activelearning
from antgo.framework.helper.exporter import *
from antgo.framework.helper.models.detectors import *
from antgo.framework.helper.dataset.pipelines import *
from antgo.framework.helper.utils import Config
from antgo.framework.helper.dataset import *
import json

# 2.step 导入自定义系统后台(包括hdfs后端,KV后端)
from system import *

# 3.step 导入扩展模块 (包括自定义的模型,数据集,,等)
from models import *
from metrics import *
from dataset import *
from hooks import *

# 4.step 定义shell参数
# 4.1.step 自定义扩展参数
# 例子：
# args.DEFINE_string('root_hdfs', '', 'label dataset hdfs')

# 4.2.step 定义标准nn参数
args.DEFINE_nn_args()

def main():
    # step1: 加载参数
    nn_args = args.parse_args()
    args.print_args(nn_args)
    assert(nn_args.exp != '')
    # step2: 加载配置文件
    if not os.path.exists(nn_args.config):
        here_dir = os.path.dirname(os.path.realpath(__file__))
        # step2.1: 查找位置1
        config_file_path = os.path.join(here_dir, 'config.py')
        if not os.path.exists(config_file_path):
            config_file_path = ''

        # step2.2: 查找位置2
        if config_file_path == '':
            config_file_path = os.path.join(here_dir, 'configs', 'config.py')
            if not os.path.exists(config_file_path):
                config_file_path = ''

        nn_args.config = config_file_path
    if nn_args.config == '':
        print('Couldnt find correct config file.')
        return

    cfg = Config.fromfile(os.path.join(os.path.dirname(__file__), nn_args.config))
    if 'checkpoint_config' in cfg:
        cfg.checkpoint_config['out_dir'] = os.path.join(cfg.checkpoint_config['out_dir'], nn_args.exp)
    if 'evaluation' in cfg:
        cfg.evaluation['out_dir'] = os.path.join(cfg.evaluation['out_dir'], nn_args.exp)

    for data_stage in ['train', 'val', 'test']:
        # 实验默认的训练，验证，测试数据
        # 如果配置的是远程数据路径，则在本地寻找对应路径数据
        if not getattr(cfg.data, data_stage, None):
            continue

    for data_stage in ['train', 'val', 'test']:
        # 实验默认的训练，验证，测试数据
        # 如果配置的是远程数据路径，则在本地寻找对应路径数据
        if not getattr(cfg.data, data_stage, None):
            continue

        if isinstance(getattr(cfg.data, data_stage), list):
            data_part_0 = getattr(cfg.data, data_stage)[0]
            data_part_1 = getattr(cfg.data, data_stage)[1]

            if not getattr(data_part_0, 'data_path_list', None):
                continue
            for i in range(len(data_part_0.data_path_list)):
                data_record_path = data_part_0.data_path_list[i]
                if data_record_path.startswith('hdfs') or data_record_path.startswith('http'):
                    try_local_path = f"dataset-storage/{data_stage}/{data_record_path.split('/')[-1]}"
                    data_part_0.data_path_list[i] = try_local_path

            if not getattr(data_part_1, 'data_path_list', None):
                continue
            for i in range(len(data_part_1.data_path_list)):                
                data_record_path = data_part_1.data_path_list[i]
                if data_record_path.startswith('hdfs') or data_record_path.startswith('http'):
                    try_local_path = f"dataset-storage/{data_stage}/{data_record_path.split('/')[-1]}"
                    data_part_1.data_path_list[i] = try_local_path
        else:
            if not getattr(getattr(cfg.data, data_stage), 'data_path_list', None):
                continue
            for i in range(len(getattr(cfg.data, data_stage).data_path_list)):
                data_record_path = getattr(cfg.data, data_stage).data_path_list[i]
                if data_record_path.startswith('hdfs') or data_record_path.startswith('http'):
                    try_local_path = f"dataset-storage/{data_stage}/{data_record_path.split('/')[-1]}"
                    getattr(cfg.data, data_stage).data_path_list[i] = try_local_path   

    # step3 检查补充配置
    if nn_args.extra_config is not None and nn_args.extra_config != '':
        if os.path.exists(nn_args.extra_config):
            with open(nn_args.extra_config, 'r') as fp:
                extra_config = Config.fromstring(fp.read(),'.json')

            # extra_config 格式为项目信息格式
            # step3.1: 数据加载 (默认TFRECORD是antgo默认标准打包格式)
            # 有标签数据,伪标签数据,无标签数据,跨域数据 文件列表
            label_local_path_list = []
            pseudo_local_path_list = []
            unlabel_local_path_list = []
            cross_domain_local_path_list = []
            if os.path.exists('dataset-storage'):
                for sub_folder_name in os.listdir('dataset-storage'):
                    if not os.path.isdir(os.path.join('dataset-storage', sub_folder_name)):
                        continue

                    if sub_folder_name == 'label':
                        file_list = []
                        for data_record_file in os.listdir(os.path.join('dataset-storage', sub_folder_name)):
                            # 同一个数据文件，存在*-index和*-tfrecord
                            finding_file = '-'.join(data_record_file.split('-')[:-1])
                            if finding_file not in file_list:
                                file_list.append(finding_file)
                        
                        for data_record_file in file_list:
                            label_local_path_list.append(os.path.join('dataset-storage', sub_folder_name, data_record_file))
                    elif sub_folder_name == 'pseudo-label':
                        file_list = []
                        for data_record_file in os.listdir(os.path.join('dataset-storage', sub_folder_name)):
                            # 同一个数据文件，存在*-index和*-tfrecord
                            finding_file = '-'.join(data_record_file.split('-')[:-1])
                            if finding_file not in file_list:
                                file_list.append(finding_file)

                        for data_record_file in file_list:
                            pseudo_local_path_list.append(os.path.join('dataset-storage', sub_folder_name, data_record_file))
                    elif sub_folder_name == 'unlabel':
                        file_list = []
                        for data_record_file in os.listdir(os.path.join('dataset-storage', sub_folder_name)):
                            # 同一个数据文件，存在*-index和*-tfrecord
                            finding_file = '-'.join(data_record_file.split('-')[:-1])
                            if finding_file not in file_list:
                                file_list.append(finding_file)

                        for data_record_file in file_list:
                            unlabel_local_path_list.append(os.path.join('dataset-storage', sub_folder_name, data_record_file))

            # step3.2: 数据使用方式扩展
            existed_train_datalist = []
            if 'data' in extra_config:
                # 扩展数据的使用方式 (训练集)
                # 训练数据的使用方式扩展
                if 'train' in extra_config['data']:
                    existed_train_datalist = cfg.data.train.data_path_list
                    cfg.data.train = extra_config['data']['train']
                if 'train_dataloader' in extra_config['data']:
                    cfg.data.train_dataloader = extra_config['data']['train_dataloader']

                # 测试数据的使用方式扩展
                if 'test' in extra_config['data']:
                    cfg.data.test = extra_config['data']['test']
                    cfg.data.test.data_path_list = unlabel_local_path_list
                if 'test_dataloader' in extra_config['data']:
                    cfg.data.test_dataloader = extra_config['data']['test_dataloader']

            # step3.3: 数据文件数扩展
            # 训练集相关
            # 功能性数据类 RepeatDataset, DatasetSamplingByClass
            control_dataset_type_names = ['RepeatDataset', 'DatasetSamplingByClass']
            if not isinstance(cfg.data.train, list):
                # 默认仅支持TFDataset
                # 默认常规训练方式
                concrete_dataset = cfg.data.train
                while True:
                    if concrete_dataset.type in control_dataset_type_names:
                        concrete_dataset = concrete_dataset.dataset
                        continue
                    concrete_dataset.data_path_list.extend(existed_train_datalist)
                    concrete_dataset.data_path_list.extend(label_local_path_list)
                    concrete_dataset.data_path_list.extend(pseudo_local_path_list)
                    break
            else:
                # 默认复杂训练方式（如，半监督训练，跨域训练模式）
                # 仅支持list size=2 （默认，0: 标签数据+伪标签数据；1: 无标签数据）
                concrete_dataset = cfg.data.train[0]
                while True:
                    if concrete_dataset.type in control_dataset_type_names:
                        concrete_dataset = concrete_dataset.dataset
                        continue
                    concrete_dataset.data_path_list.extend(existed_train_datalist)
                    concrete_dataset.data_path_list.extend(label_local_path_list)
                    concrete_dataset.data_path_list.extend(pseudo_local_path_list)
                    break

                concrete_dataset = cfg.data.train[1]
                while True:
                    if concrete_dataset.type in control_dataset_type_names:
                        concrete_dataset = concrete_dataset.dataset
                        continue
                    # 无标签数据
                    concrete_dataset.data_path_list.extend(unlabel_local_path_list)
                    # 跨域数据
                    concrete_dataset.data_path_list.extend(cross_domain_local_path_list)
                    break

            # 测试集相关 (仅存在无标签数据时，默认是主动学习过程)
            if nn_args.process == 'activelearning':
                assert(len(unlabel_local_path_list) > 0)
                cfg.data.test.data_path_list = unlabel_local_path_list

            # step3.4: 模型配置扩展
            if 'model' in extra_config:
                # 更新默认配置
                default_model_cfg = cfg.model
                if 'model' in extra_config['model'] and \
                    'teacher' in extra_config['model']['model']:
                    cfg.model = dict(
                        type=extra_config['model']['type'],
                        model=dict(
                            teacher=extra_config['model']['model']['teacher'],
                            student=default_model_cfg
                        ),
                        train_cfg=extra_config['model']['train_cfg'] if 'train_cfg' in extra_config['model'] else None,
                        test_cfg=extra_config['model']['test_cfg'] if 'test_cfg' in extra_config['model'] else None,
                        init_cfg=extra_config['model']['init_cfg'] if 'init_cfg' in extra_config['model'] else None
                    )
                    
                    if cfg.model.model.teacher is None:
                        cfg.model.model.teacher = default_model_cfg
                    if cfg.model.model.student is None:
                        cfg.model.model.student = default_model_cfg
                        
                    if cfg.model.train_cfg is not None:
                        if 'student' in cfg.model.train_cfg:
                            align_name = cfg.model.train_cfg.align
                            if 'channels' in cfg.model.train_cfg.student:
                                cfg.model.train_cfg.student.channels = cfg.info.get(align_name).channels
                            if 'shapes' in cfg.model.train_cfg.student:
                                cfg.model.train_cfg.student.shapes = cfg.info.get(align_name).shapes 
                else:
                    cfg.model = dict(
                        type=extra_config['model']['type'],
                        model=default_model_cfg,
                        train_cfg=extra_config['model']['train_cfg'] if 'train_cfg' in extra_config['model'] else None,
                        test_cfg=extra_config['model']['test_cfg'] if 'test_cfg' in extra_config['model'] else None,
                        init_cfg=extra_config['model']['init_cfg'] if 'init_cfg' in extra_config['model'] else None
                    )

            # step3.5: 模型运行时hooks扩展
            if 'custom_hooks' in extra_config:
                hooks = extra_config['custom_hooks']
                if not isinstance(hooks, list):
                    hooks = [hooks]
                
                # 更新默认配置
                custom_hooks = getattr(cfg, 'custom_hooks', None)
                if custom_hooks is None:
                    custom_hooks = hooks
                else:
                    if isinstance(custom_hooks, list):
                        custom_hooks.extend(hooks)
                    else:
                        custom_hooks = hooks
            
            # step3.6: 模型其他方面扩展
            if 'optimizer' in extra_config:
                cfg.optimizer = extra_config['optimizer']
            if 'optimizer_config' in extra_config:
                cfg.optimizer_config = extra_config['optimizer_config']
            if 'lr_config' in extra_config:
                cfg.lr_config = extra_config['lr_config']
            if 'max_epochs' in extra_config:
                cfg.max_epochs = extra_config['max_epochs']

    # step4 检查checkpoint
    if nn_args.checkpoint is not None and nn_args.checkpoint != '':
        if not nn_args.checkpoint.startswith('/') and not nn_args.checkpoint.startswith('./'):
            checkpoint_file_name = nn_args.checkpoint.split('/')[-1]
            nn_args.checkpoint = os.path.join('checkpoint-storage', checkpoint_file_name)
            if not os.path.exists(nn_args.checkpoint):
                logging.error(f'Checkpoint {nn_args.checkpoint} not in local.')
    if nn_args.resume_from is not None and nn_args.resume_from != '':
        if not nn_args.resume_from.startswith('/') and not nn_args.resume_from.startswith('./'):
            checkpoint_file_name = nn_args.resume_from.split('/')[-1]
            nn_args.resume_from = os.path.join('checkpoint-storage', checkpoint_file_name)
            if not os.path.exists(nn_args.resume_from):
                logging.error(f'Checkpoint {nn_args.resume_from} not in local.')

    # step5 添加root (运行时，输出结果保存的根目录地址)
    cfg.root = nn_args.root if nn_args.root != '' else './output/'
    cfg.root = os.path.join(cfg.root, nn_args.exp)
    # step5.1 添加root地址（影响checkpoint_config, evaluation）
    if cfg.root != '':
        cfg.checkpoint_config.out_dir = cfg.root
        cfg.evaluation.out_dir = cfg.root
    
    # step6: 执行指令(训练、测试、模型导出)
    if nn_args.process == 'train':
        # 创建训练过程
        trainer = Trainer(
            cfg, 
            './', 
            int(nn_args.gpu_id), # 对于多卡运行环境,会自动忽略此参数数值
            distributed=nn_args.distributed, 
            diff_seed=nn_args.diff_seed, 
            deterministic=nn_args.deterministic, 
            find_unused_parameters=nn_args.find_unused_parameters)

        trainer.config_dataloader(with_validate=not nn_args.no_validate)
        trainer.config_model(resume_from=nn_args.resume_from, load_from=nn_args.checkpoint)
        if nn_args.max_epochs < 0:
            # 优先使用外部指定max_epochs
            nn_args.max_epochs = cfg.max_epochs
        print(f'max epochs {nn_args.max_epochs}')

        trainer.start_train(max_epochs=nn_args.max_epochs)
    elif nn_args.process == 'test':
        # 创建测试过程
        print(f'nn_args.distributed {nn_args.distributed}')
        print(f'nn_args.gpu-id {nn_args.gpu_id}')
        print(f'nn_args.checkpoint {nn_args.checkpoint}')
        tester = Tester(
            cfg, 
            './',
            int(nn_args.gpu_id), # 对于多卡运行环境,会自动忽略此参数数值
            distributed=nn_args.distributed)
        tester.config_model(checkpoint=nn_args.checkpoint)
        tester.evaluate()
    elif nn_args.process == 'activelearning':
        # 创建主动学习过程,挑选等待标注样本
        print(f'nn_args.distributed {nn_args.distributed}')
        print(f'nn_args.gpu-id {nn_args.gpu_id}')
        print(f'nn_args.checkpoint {nn_args.checkpoint}')        
        ac = Activelearning(cfg, './', int(nn_args.gpu_id), distributed=nn_args.distributed)
        ac.config_model(checkpoint=nn_args.checkpoint, revise_keys=[('^','model.')])
        ac.select(nn_args.exp)
    elif nn_args.process == 'export':
        # 创建导出模型过程
        tester = Exporter(cfg, './')
        checkpoint_file_name = nn_args.checkpoint.split('/')[-1].split('.')[0]
        tester.export(
            input_tensor_list=[torch.zeros(shape, dtype=torch.float32) for shape in cfg.export.input_shape_list], 
            input_name_list=cfg.export.input_name_list, 
            output_name_list=cfg.export.output_name_list, 
            checkpoint=nn_args.checkpoint, 
            prefix=f'{nn_args.exp}-{checkpoint_file_name}-model')


if __name__ == "__main__":
    main()


# -*- coding: UTF-8 -*-
# @Time    : 2022/9/3 00:54
# @File    : activelearning_v2.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import time
import cv2
from antgo.resource.html import *
from antgo.ant.base import *
from antgo.ant.base import _pick_idle_port
from antgo.dataflow.common import *
from antgo.dataflow.recorder import *
from antvis.client.httprpc import *
from multiprocessing import Process
from antgo.task.task import *
from antgo.crowdsource.label_server import *
import traceback
import subprocess
import os
import socket
import requests
import json
import zipfile
import imagesize


class ActivelearningDataRecorder(object):
  def __init__(self, rpc, dump_dir):
    self.rpc = rpc
    self.activelearning_static_dir = os.path.join(dump_dir, 'static')

  def record(self, sample):
    if not os.path.exists(os.path.join(self.activelearning_static_dir, 'data')):
      os.makedirs(os.path.join(self.activelearning_static_dir, 'data'))

    assert('image_file' in sample or 'id' in sample)
    image_file = ''
    image_height = 0
    image_width = 0
    if 'image_file' in sample:
      # 将文件上传至目录
      shutil.copy(sample['image_file'], os.path.join(self.activelearning_static_dir, 'data'))
      image_file = sample['image_file']
      image_file = f'/static/data/{image_file}'
      image_width, image_height = imagesize.get(sample['image_file'])
    else:
      data = sample['image']
      if len(data.shape) == 3:
        assert (data.shape[2] == 3 or data.shape[2] == 4)

      assert (len(data.shape) == 2 or len(data.shape) == 3)

      if data.dtype != np.uint8:
        data_min = np.min(data)
        data_max = np.max(data)
        data = ((data - data_min) / (data_max - data_min) * 255).astype(np.uint8)

      sample.pop('image')
      sample_id = sample['id']
      cv2.imwrite(os.path.join(self.activelearning_static_dir, 'data', f'{sample_id}.png'), data)
      image_file = f'/static/data/{sample_id}.png'
      image_height, image_width = data.shape[:2]

    data = {}
    data['image_file'] = image_file
    data['label_info'] = sample.get('label_info', [])
    data['width'] = image_width
    data['height'] = image_height
    samples = json.dumps([data])
    # 更新待标注样本
    self.rpc.label.sample.fresh.post(samples=samples)

  def close(self):
    pass


class AntActiveLearningV2(AntBase):
  def __init__(self,
               ant_context,
               ant_name,
               data_factory,
               ant_dump_dir,
               ant_token,
               ant_task_config=None,
               **kwargs):
    super(AntActiveLearningV2, self).__init__(ant_name, ant_context, ant_token, **kwargs)
    self.ant_data_source = data_factory
    self.ant_dump_dir = ant_dump_dir
    self.ant_context.ant = self
    self.ant_task_config = ant_task_config
    self.host_ip = self.context.params.system.get('ip', '127.0.0.1')
    self.host_port = self.context.params.system.get('port', -1)
    self.rpc = None
    self._running_dataset = None
    self._running_task = None
    self.p = None
    self._round = -1

  @property
  def running_dataset(self):
    return self._running_dataset

  @property
  def running_task(self):
      return self._running_task

  def ping_until_ok(self):
    while True:
      content = self.rpc.ping.get()
      if content['status'] != 'ERROR':
        break
      # 暂停5秒钟，再进行尝试
      time.sleep(5)

  def wait_until_stop(self):
    self.p.join()

  def labeling(self):
    # 启动新一轮标注
    response = self.rpc.info.get()
    if response['content']['project_state']['stage'] == 'labeling':
      logger.error(f'Label Round {self._round} is not finish.')
      return

    self._round += 1
    self.rpc.info.post(running_state='running',  running_stage='labeling', running_round=self._round)

  def waiting(self):
    # 完成本轮标注，后台会清空标注信息
    self.rpc.info.post(running_state='running',  running_stage='waiting', running_round=self._round)
    
  def state(self):
    response = self.rpc.info.get()
    return response['content']['project_state']   

  def download(self, waiting_finish=True):
    if waiting_finish:
      # 需要等待本轮标准完成
      while True:
        response = self.rpc.info.get()
        if response['status'] == 'ERROR':
          print('rpc error...')
          time.sleep(5)
          continue

        if response['content']['project_state']['stage'] == 'finish':
          break
        # 等待10分钟后检查
        time.sleep(10)
        
    if not os.path.exists(os.path.join(self.ant_dump_dir, 'label', f'{self._round}')):
      os.makedirs(os.path.join(self.ant_dump_dir, 'label', f'{self._round}'))

    folder = os.path.join(self.ant_dump_dir, 'label', f'{self._round}')
    file_name = f'label.json'
    self.rpc.label.export.download(
      file_folder=folder,
      file_name=file_name
    )    
    with open(os.path.join(folder, file_name), 'r') as fp:
      content = json.load(fp)

      return content

  def start(self, **kwargs):
    # 0.step loading challenge task
    running_ant_task = None
    if self.token is not None:
      # 0.step load challenge task
      response = self.context.dashboard.challenge.get(command=type(self).__name__)
      if response['status'] is None:
        # invalid token
        logger.error('Couldnt load challenge task.')
        self.token = None
      elif response['status'] == 'SUSPEND':
        # prohibit submit challenge task frequently
        # submit only one in one week
        logger.error('Prohibit submit challenge task frequently.')
        exit(-1)
      elif response['status'] == 'OK':
        # maybe user token or task token
        content = response['content']
        if 'task' in content:
          challenge_task = create_task_from_json(content)
          if challenge_task is None:
            logger.error('Couldnt load challenge task.')
            exit(-1)
          running_ant_task = challenge_task
      else:
        # unknow error
        logger.error('Unknow error.')
        exit(-1)

    if running_ant_task is None:
      # 0.step load custom task
      custom_task = create_task_from_xml(self.ant_task_config, self.context)
      if custom_task is None:
        logger.error('Couldnt load custom task.')
        exit(0)
      running_ant_task = custom_task

    assert (running_ant_task is not None)
    self._running_task = running_ant_task

    #获得实验ID
    experiment_uuid = self.context.experiment_uuid
    logger.info('Build experiment folder.')
    if not os.path.exists(os.path.join(self.ant_dump_dir, experiment_uuid)):
      os.makedirs(os.path.join(self.ant_dump_dir, experiment_uuid))

    is_launch_web_server = True
    if self.host_port < 0:
      is_launch_web_server = False

    self.rpc = HttpRpc("v1", "antgo/api", self.host_ip, self.host_port)
    self.context.recorder = \
      ActivelearningDataRecorder(
        self.rpc,
        os.path.join(self.ant_dump_dir, 'activelearning')
      )

    if is_launch_web_server:
      # 准备web服务资源
      static_folder = '/'.join(os.path.dirname(__file__).split('/')[0:-1])
      activelearning_static_dir = os.path.join(self.ant_dump_dir, 'activelearning', 'static')
      if os.path.exists(activelearning_static_dir):
        shutil.rmtree(activelearning_static_dir)

      shutil.copytree(os.path.join(static_folder, 'resource', 'app'), activelearning_static_dir)

      # 准备有效端口
      self.host_port = _pick_idle_port(self.host_port)

      # 在独立进程中启动webserver
      sample_metas = {'filters': []}
      metas = self.context.params.activelearning.metas.get()
      white_users = self.context.params.activelearning.white_users.get() \
        if self.context.params.activelearning.white_users is not None else None
      task_metas = {
        'task_type': self.running_task.task_type,
        'task_name': self.running_task.task_name,
        'label_type': self.context.params.activelearning.label_type
      }
      
      data_json_file = kwargs.get('json_file', None)
      sample_folder = None
      sample_list = []
      if data_json_file is not None:
        # 直接使用来自于data_json_file中的样本
        with open(data_json_file, 'r') as fp:
          sample_list = json.load(fp)      
        sample_folder = os.path.dirname(data_json_file)
        
      self.p = \
        multiprocessing.Process(
          target=label_server_start,
          args=(os.path.join(self.ant_dump_dir, 'activelearning'),
                self.host_port,
                task_metas,
                sample_metas,
                metas,
                sample_folder,
                sample_list,
                {
                  'state': 'running', 
                  'stage': 'waiting'    # 标注服务启动后，处在等待状态。之后数据推送好后，重制状态使其处在标注状态
                },
                white_users)
        )
      self.p.daemon = True  # 主进程结束后，自进程也将结束
      self.p.start()

      # 等待直到http服务开启
      self.ping_until_ok()

    if self.context.is_interact_mode:
      logger.info('Running on interact mode.')
      return

    # TODO，
    # 对于非交互模式，设置训练->测试->采样->标注，全局调度
    self.p.join()



from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
import math
from ...builder import DISTILL_LOSSES

'''
In the original paper, AFD is one of components of AFDS.
AFDS: Attention Feature Distillation and Selection
AFD:  Attention Feature Distillation
AFS:  Attention Feature Selection
We find the original implementation of attention is unstable, thus we replace it with a SE block.
'''


@DISTILL_LOSSES.register_module()
class AFD(nn.Module):
	'''
	Pay Attention to Features, Transfer Learn Faster CNNs
	https://openreview.net/pdf?id=ryxyCeHtPB
	'''
	def __init__(self, in_channels, att_f):
		super(AFD, self).__init__()
		mid_channels = int(in_channels * att_f)

		self.attention = nn.Sequential(*[
				nn.Conv2d(in_channels, mid_channels, 1, 1, 0, bias=True),
				nn.ReLU(inplace=True),
				nn.Conv2d(mid_channels, in_channels, 1, 1, 0, bias=True)
			])

		for m in self.modules():
			if isinstance(m, nn.Conv2d):
				nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
				if m.bias is not None:
					nn.init.constant_(m.bias, 0)
		
	def forward(self, fm_s, fm_t, eps=1e-6):
		fm_t_pooled = F.adaptive_avg_pool2d(fm_t, 1)
		rho = self.attention(fm_t_pooled)
		# rho = F.softmax(rho.squeeze(), dim=-1)
		rho = torch.sigmoid(rho.squeeze())
		rho = rho / torch.sum(rho, dim=1, keepdim=True)

		fm_s_norm = torch.norm(fm_s, dim=(2,3), keepdim=True)
		fm_s      = torch.div(fm_s, fm_s_norm+eps)
		fm_t_norm = torch.norm(fm_t, dim=(2,3), keepdim=True)
		fm_t      = torch.div(fm_t, fm_t_norm+eps)

		loss = rho * torch.pow(fm_s-fm_t, 2).mean(dim=(2,3))
		loss = loss.sum(1).mean(0)

		return loss

# class AFD(nn.Module):
# 	'''
# 	Pay Attention to Features, Transfer Learn Faster CNNs
# 	https://openreview.net/pdf?id=ryxyCeHtPB
# 	'''
# 	def __init__(self, chw):
# 		super(AFD, self).__init__()
# 		c, h, w = chw

# 		self.weight1 = nn.Parameter(math.sqrt(2.0) / math.sqrt(h*w) * torch.randn(h, h*w))
# 		self.bias1   = nn.Parameter(torch.zeros(h))

# 		self.weight2 = nn.Parameter(math.sqrt(2.0) / math.sqrt(h) * torch.randn(h))
# 		self.bias2   = nn.Parameter(torch.zeros(c))

# 	def forward(self, fm_s, fm_t, eps=1e-6):
# 		b, c, h, w = fm_t.size()

# 		fm_t_flatten = fm_t.view(fm_t.size(0), fm_t.size(1), -1)
# 		weight1 = torch.stack([self.weight1.t()]*b, dim=0)
# 		bias1   = self.bias1.unsqueeze(0).unsqueeze(1)
# 		rho     = F.relu(torch.bmm(fm_t_flatten, weight1) + bias1)
# 		weight2 = self.weight2.view(-1, 1)
# 		bias2   = self.bias2.unsqueeze(0)
# 		rho     = torch.mm(rho.view(-1, rho.size(2)), weight2).view(b,c) + bias2
# 		# rho     = F.softmax(rho, dim=-1)
# 		rho = torch.sigmoid(rho)
# 		rho = rho / torch.sum(rho, dim=1, keepdim=True)
# 		# print(rho)

# 		fm_s_norm = torch.norm(fm_s, dim=(2,3), keepdim=True)
# 		fm_s      = torch.div(fm_s, fm_s_norm+eps)
# 		fm_t_norm = torch.norm(fm_t, dim=(2,3), keepdim=True)
# 		fm_t      = torch.div(fm_t, fm_t_norm+eps)

# 		loss = rho * torch.pow(fm_s-fm_t, 2).mean(dim=(2,3))
# 		loss = loss.sum(1).mean(0)

# 		return loss


# -*- coding: UTF-8 -*-
# @Time    : 2019-06-08 15:51
# @File    : visalso.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
import sys
from antgo.dataflow.dataset import *
import scipy.io as io
import os
import numpy as np
import cv2
import time

__all__ = ['VisalSO']

class VisalSO(Dataset):
    def __init__(self, train_or_test="train", dir=None, ext_params=None):
        super().__init__(train_or_test, dir, ext_params)
        # 不区分训练集和测试集，仅用来进行模型正确性测试        
        url_address = 'http://visal.cs.cityu.edu.hk/static/downloads/SmallObjectDataset.zip'

        self.class_name = [
            'fish',
            'fly',
            'honeybee',
            'seagull'
        ]

        if os.environ.get('LOCAL_RANK', 0) == 0:
            if not os.path.exists(os.path.join(self.dir, 'Small Object dataset')):
                # 数据集不存在，需要重新下载，并创建标记
                if not os.path.exists(self.dir):
                    os.makedirs(self.dir)
    
                os.system(f'cd {self.dir} && wget {url_address} && unzip SmallObjectDataset.zip')
                os.system('touch DATASET_IS_READY')
            else:
                # 数据集存在，创建标记
                if not os.path.exists('DATASET_IS_READY'):
                    os.system('touch DATASET_IS_READY')
        else:
            while True:
                # 等待直到存在指定文件
                if os.path.exists('DATASET_IS_READY'):
                    break
                time.sleep(5)

        self.image_file_list = []
        self.bboxes_list = []
        for ci, cn in enumerate(self.class_name):
            subfolder = os.path.join(self.dir, 'Small Object dataset', cn)
            cn_gt_bbox_folder = os.path.join(subfolder, 'gt-bbox')
            cn_img_folder = os.path.join(subfolder, 'img')

            image_id_to_file_name_map = {}
            for image_file_name in os.listdir(cn_img_folder):
                image_id = image_file_name.split('.')[0][-3:]
                image_id_to_file_name_map[image_id] = image_file_name

            for mat_file_name in os.listdir(cn_gt_bbox_folder):
                matr = io.loadmat(os.path.join(cn_gt_bbox_folder, mat_file_name))
                bboxes = matr['bbox_all'].astype(np.int32)
                x0 = bboxes[:,0] - bboxes[:,2]/2
                y0 = bboxes[:,1] - bboxes[:,3]/2
                x1 = bboxes[:,0] + bboxes[:,2]/2
                y1 = bboxes[:,1] + bboxes[:,3]/2

                bboxes[:,0] = x0
                bboxes[:,1] = y0
                bboxes[:,2] = x1
                bboxes[:,3] = y1
                self.bboxes_list.append((bboxes, np.array([ci for _ in range(len(bboxes))])))

                image_id = mat_file_name.split('.')[0].split('_')[-1]
                image_file_name = image_id_to_file_name_map[image_id]
                self.image_file_list.append(os.path.join(cn_img_folder, image_file_name))

    @property
    def size(self):
        return len(self.image_file_list)
    
    def at(self, id):
        image = cv2.imread(self.image_file_list[id])
        bboxes, labels = self.bboxes_list[id]
        return (image, {'bboxes': bboxes, 'labels': labels, 'image_meta': {'image_shape': (image.shape[0], image.shape[1])}})
    
    def split(self, split_params={}, split_method=''):
        raise NotImplementedError

# vso = VisalSO('train', './visalso_dataset')
# for i in range(4):
#     info = vso.sample(i)
    
#     image = info['image']
#     gt_bbox = info['bboxes']
#     for x0,y0,x1,y1 in gt_bbox:
#         image = cv2.rectangle(image, (int(x0), int(y0)), (int(x1), int(y1)) ,(0,0,255), 1)
        
#     cv2.imwrite('/root/workspace/handtracking/visalso_dataset/abcd.png', image)    
#     print('d')


from . import backend
from .common import register, register_cstruct, CUSTOM_OP_LIST
from . import common
import ctypes
common.backend = backend

class CFTensor(ctypes.Structure):
    def __init__(self, data):
        self.val = data
        super().__init__(ctypes.c_size_t(len(data.shape)), ctypes.cast((ctypes.c_size_t * len(data.shape))(*data.shape), ctypes.POINTER(ctypes.c_size_t)), ctypes.cast(data.ctypes.data, ctypes.POINTER(ctypes.c_float)), False)

    _fields_ = [("dim_size", ctypes.c_size_t), ("dims", ctypes.POINTER(ctypes.c_size_t)), ("data", ctypes.POINTER(ctypes.c_float)), ('is_assign_inner', ctypes.c_bool)]

class CITensor(ctypes.Structure):
    def __init__(self, data):
        self.val = data
        super().__init__(ctypes.c_size_t(len(data.shape)), ctypes.cast((ctypes.c_size_t * len(data.shape))(*data.shape), ctypes.POINTER(ctypes.c_size_t)), ctypes.cast(data.ctypes.data, ctypes.POINTER(ctypes.c_int)), False)

    _fields_ = [("dim_size", ctypes.c_size_t), ("dims", ctypes.POINTER(ctypes.c_size_t)), ("data", ctypes.POINTER(ctypes.c_int)),  ('is_assign_inner', ctypes.c_bool)]

class CUCTensor(ctypes.Structure):
    def __init__(self, data):
        self.val = data
        super().__init__(ctypes.c_size_t(len(data.shape)), ctypes.cast((ctypes.c_size_t * len(data.shape))(*data.shape), ctypes.POINTER(ctypes.c_size_t)), ctypes.cast(data.ctypes.data, ctypes.POINTER(ctypes.c_ubyte)), False)

    _fields_ = [("dim_size", ctypes.c_size_t), ("dims", ctypes.POINTER(ctypes.c_size_t)), ("data", ctypes.POINTER(ctypes.c_ubyte)),  ('is_assign_inner', ctypes.c_bool)]

class CDTensor(ctypes.Structure):
    def __init__(self, data):
        self.val = data
        super().__init__(ctypes.c_size_t(len(data.shape)), ctypes.cast((ctypes.c_size_t * len(data.shape))(*data.shape), ctypes.POINTER(ctypes.c_size_t)), ctypes.cast(data.ctypes.data, ctypes.POINTER(ctypes.c_double)), False)

    _fields_ = [("dim_size", ctypes.c_size_t), ("dims", ctypes.POINTER(ctypes.c_size_t)), ("data", ctypes.POINTER(ctypes.c_double)), ('is_assign_inner', ctypes.c_bool)]

common.register_cstruct('CDTensor', CDTensor)
common.register_cstruct('CFTensor', CFTensor)
common.register_cstruct('CITensor', CITensor)
common.register_cstruct('CUCTensor', CUCTensor)



import torch
import torch.nn as nn

from antgo.framework.helper.models.builder import LOSSES
from .utils import weighted_loss


@weighted_loss
def smooth_l1_loss(pred, target, beta=1.0):
    """Smooth L1 loss.

    Args:
        pred (torch.Tensor): The prediction.
        target (torch.Tensor): The learning target of the prediction.
        beta (float, optional): The threshold in the piecewise function.
            Defaults to 1.0.

    Returns:
        torch.Tensor: Calculated loss
    """
    assert beta > 0
    if target.numel() == 0:
        return pred.sum() * 0

    assert pred.size() == target.size()
    diff = torch.abs(pred - target)
    loss = torch.where(diff < beta, 0.5 * diff * diff / beta,
                       diff - 0.5 * beta)
    return loss


@weighted_loss
def l1_loss(pred, target):
    """L1 loss.

    Args:
        pred (torch.Tensor): The prediction.
        target (torch.Tensor): The learning target of the prediction.

    Returns:
        torch.Tensor: Calculated loss
    """
    if target.numel() == 0:
        return pred.sum() * 0

    assert pred.size() == target.size()
    loss = torch.abs(pred - target)
    return loss


@LOSSES.register_module()
class SmoothL1Loss(nn.Module):
    """Smooth L1 loss.

    Args:
        beta (float, optional): The threshold in the piecewise function.
            Defaults to 1.0.
        reduction (str, optional): The method to reduce the loss.
            Options are "none", "mean" and "sum". Defaults to "mean".
        loss_weight (float, optional): The weight of loss.
    """

    def __init__(self, beta=1.0, reduction='mean', loss_weight=1.0):
        super(SmoothL1Loss, self).__init__()
        self.beta = beta
        self.reduction = reduction
        self.loss_weight = loss_weight

    def forward(self,
                pred,
                target,
                weight=None,
                avg_factor=None,
                reduction_override=None,
                **kwargs):
        """Forward function.

        Args:
            pred (torch.Tensor): The prediction.
            target (torch.Tensor): The learning target of the prediction.
            weight (torch.Tensor, optional): The weight of loss for each
                prediction. Defaults to None.
            avg_factor (int, optional): Average factor that is used to average
                the loss. Defaults to None.
            reduction_override (str, optional): The reduction method used to
                override the original reduction method of the loss.
                Defaults to None.
        """
        assert reduction_override in (None, 'none', 'mean', 'sum')
        reduction = (
            reduction_override if reduction_override else self.reduction)
        loss_bbox = self.loss_weight * smooth_l1_loss(
            pred,
            target,
            weight,
            beta=self.beta,
            reduction=reduction,
            avg_factor=avg_factor,
            **kwargs)
        return loss_bbox


@LOSSES.register_module()
class L1Loss(nn.Module):
    """L1 loss.

    Args:
        reduction (str, optional): The method to reduce the loss.
            Options are "none", "mean" and "sum".
        loss_weight (float, optional): The weight of loss.
    """

    def __init__(self, reduction='mean', loss_weight=1.0):
        super(L1Loss, self).__init__()
        self.reduction = reduction
        self.loss_weight = loss_weight

    def forward(self,
                pred,
                target,
                weight=None,
                avg_factor=None,
                reduction_override=None):
        """Forward function.

        Args:
            pred (torch.Tensor): The prediction.
            target (torch.Tensor): The learning target of the prediction.
            weight (torch.Tensor, optional): The weight of loss for each
                prediction. Defaults to None.
            avg_factor (int, optional): Average factor that is used to average
                the loss. Defaults to None.
            reduction_override (str, optional): The reduction method used to
                override the original reduction method of the loss.
                Defaults to None.
        """
        assert reduction_override in (None, 'none', 'mean', 'sum')
        reduction = (
            reduction_override if reduction_override else self.reduction)
        loss_bbox = self.loss_weight * l1_loss(
            pred, target, weight, reduction=reduction, avg_factor=avg_factor)
        return loss_bbox


@LOSSES.register_module()
class L1LossMask(nn.Module):
    def __init__(self, reduction='mean', loss_weight=1.0, is_bbox=True):
        super(L1LossMask, self).__init__()
        self.reduction = reduction
        self.loss_weight = loss_weight
        self.is_bbox = is_bbox

    def forward(self,
                pred,
                target,
                indices,
                mask):
        b, c, _, _ = pred.shape
        k = target.shape[-1]  # k, the num of obj_max
        pred = pred.reshape(b, c, -1).permute(0, 2, 1)  # (b,2,h*w) -> (b,h*w,2)
        indices = indices.long()
        mid_y_pred = torch.zeros([b, k] + list(pred.size()[2:])).to(target.device)  # (b,2,2)
        for i in range(mid_y_pred.size(0)):
            mid_y_pred[i] = pred[i, indices[i], :]
        mid_y_pred = mid_y_pred.permute(0, 2, 1)  # (b, 2, 2)
        mask = mask.reshape((b, 1, -1)).repeat(1, c, 1)
        if self.is_bbox:
            mask = mask.permute(0, 2, 1)
            target = torch.squeeze(target)
        total_loss = torch.sum(torch.abs(target * mask - mid_y_pred * mask))
        if self.reduction in ['mean', 'none', None]:
            loss = total_loss / (torch.sum(mask) + 1e-5)
        else:
            loss = total_loss
        return loss * self.loss_weight



import numpy as np
import torch
import torch.nn as nn

from antgo.framework.helper.models.builder import LOSSES
from .utils import weighted_loss


@weighted_loss
def balanced_l1_loss(pred,
                     target,
                     beta=1.0,
                     alpha=0.5,
                     gamma=1.5,
                     reduction='mean'):
    """Calculate balanced L1 loss.

    Please see the `Libra R-CNN <https://arxiv.org/pdf/1904.02701.pdf>`_

    Args:
        pred (torch.Tensor): The prediction with shape (N, 4).
        target (torch.Tensor): The learning target of the prediction with
            shape (N, 4).
        beta (float): The loss is a piecewise function of prediction and target
            and ``beta`` serves as a threshold for the difference between the
            prediction and target. Defaults to 1.0.
        alpha (float): The denominator ``alpha`` in the balanced L1 loss.
            Defaults to 0.5.
        gamma (float): The ``gamma`` in the balanced L1 loss.
            Defaults to 1.5.
        reduction (str, optional): The method that reduces the loss to a
            scalar. Options are "none", "mean" and "sum".

    Returns:
        torch.Tensor: The calculated loss
    """
    assert beta > 0
    if target.numel() == 0:
        return pred.sum() * 0

    assert pred.size() == target.size()

    diff = torch.abs(pred - target)
    b = np.e**(gamma / alpha) - 1
    loss = torch.where(
        diff < beta, alpha / b *
        (b * diff + 1) * torch.log(b * diff / beta + 1) - alpha * diff,
        gamma * diff + gamma / b - alpha * beta)

    return loss


@LOSSES.register_module()
class BalancedL1Loss(nn.Module):
    """Balanced L1 Loss.

    arXiv: https://arxiv.org/pdf/1904.02701.pdf (CVPR 2019)

    Args:
        alpha (float): The denominator ``alpha`` in the balanced L1 loss.
            Defaults to 0.5.
        gamma (float): The ``gamma`` in the balanced L1 loss. Defaults to 1.5.
        beta (float, optional): The loss is a piecewise function of prediction
            and target. ``beta`` serves as a threshold for the difference
            between the prediction and target. Defaults to 1.0.
        reduction (str, optional): The method that reduces the loss to a
            scalar. Options are "none", "mean" and "sum".
        loss_weight (float, optional): The weight of the loss. Defaults to 1.0
    """

    def __init__(self,
                 alpha=0.5,
                 gamma=1.5,
                 beta=1.0,
                 reduction='mean',
                 loss_weight=1.0):
        super(BalancedL1Loss, self).__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.beta = beta
        self.reduction = reduction
        self.loss_weight = loss_weight

    def forward(self,
                pred,
                target,
                weight=None,
                avg_factor=None,
                reduction_override=None,
                **kwargs):
        """Forward function of loss.

        Args:
            pred (torch.Tensor): The prediction with shape (N, 4).
            target (torch.Tensor): The learning target of the prediction with
                shape (N, 4).
            weight (torch.Tensor, optional): Sample-wise loss weight with
                shape (N, ).
            avg_factor (int, optional): Average factor that is used to average
                the loss. Defaults to None.
            reduction_override (str, optional): The reduction method used to
                override the original reduction method of the loss.
                Options are "none", "mean" and "sum".

        Returns:
            torch.Tensor: The calculated loss
        """
        assert reduction_override in (None, 'none', 'mean', 'sum')
        reduction = (
            reduction_override if reduction_override else self.reduction)
        loss_bbox = self.loss_weight * balanced_l1_loss(
            pred,
            target,
            weight,
            alpha=self.alpha,
            gamma=self.gamma,
            beta=self.beta,
            reduction=reduction,
            avg_factor=avg_factor,
            **kwargs)
        return loss_bbox



from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

class HDFSClient(object):
    def get(self, remote_path, local_path):
        raise NotImplementedError

    def rm(self, remote_path, is_folder=False):
        raise NotImplementedError

    def ls(self, remote_path):
        raise NotImplementedError

    def mkdir(self, remote_path, p=False):
        raise NotImplementedError

    def put(self, remote_path, local_path, is_exist=False):
        raise NotImplementedError

    def mv(self, remote_src_path, remote_target_path):
        raise NotImplementedError

    def exists(self, remote_path):
        raise NotImplementedError


class DummyKVReader(object):
    def __init__(self, *args, **kwargs):
        pass
    
    def read_many(self, index):
        pass


class DummyKVWriter(object):
    def __init__(self, *args, **kwargs):
        pass
    def write_many(self, keys, values):
        pass
    def flush(self):
        pass


hdfs_client = HDFSClient()
KVReader = DummyKVReader
KVWriter = DummyKVWriter



import warnings
from collections import Counter, Mapping, Sequence
from numbers import Number
from typing import Dict, List

import numpy as np
import torch
from antgo.framework.helper.models.utils.structures import BitmapMasks
from torch.nn import functional as F

_step_counter = Counter()


def list_concat(data_list: List[list]):
    if isinstance(data_list[0], torch.Tensor):
        return torch.cat(data_list)
    else:
        endpoint = [d for d in data_list[0]]

        for i in range(1, len(data_list)):
            endpoint.extend(data_list[i])
        return endpoint

def list_stack(data_list):
    if isinstance(data_list[0], torch.Tensor) and len(data_list[0].shape) > 2:
        return torch.stack(data_list)
    elif isinstance(data_list[0], torch.Tensor):
        return data_list
    elif isinstance(data_list[0], dict):
        return data_list
    else:
        endpoint = [d for d in data_list[0]]

        for i in range(1, len(data_list)):
            endpoint.extend(data_list[i])
        return endpoint

def sequence_concat(a, b):
    if isinstance(a, Sequence) and isinstance(b, Sequence):
        return a + b
    else:
        return None


def dict_concat(dicts: List[Dict[str, list]]):
    return {k: list_concat([d[k] for d in dicts]) for k in dicts[0].keys()}


def dict_stack(dicts, keys):
    return {k: list_stack([d[k] for d in dicts]) for k in keys}


def dict_fuse(obj_list, reference_obj):
    if isinstance(reference_obj, torch.Tensor):
        return torch.stack(obj_list)
    return obj_list


def dict_select(dict1: Dict[str, list], key: str, value: str):
    flag = [v == value for v in dict1[key]]
    return {
        k: dict_fuse([vv for vv, ff in zip(v, flag) if ff], v) for k, v in dict1.items()
    }


def dict_split(dict1, key):
    group_names = list(set(dict1[key]))
    dict_groups = {k: dict_select(dict1, key, k) for k in group_names}

    return dict_groups


def dict_sum(a, b):
    if isinstance(a, dict):
        assert isinstance(b, dict)
        return {k: dict_sum(v, b[k]) for k, v in a.items()}
    elif isinstance(a, list):
        assert len(a) == len(b)
        return [dict_sum(aa, bb) for aa, bb in zip(a, b)]
    else:
        return a + b


def zero_like(tensor_pack, prefix=""):
    if isinstance(tensor_pack, Sequence):
        return [zero_like(t) for t in tensor_pack]
    elif isinstance(tensor_pack, Mapping):
        return {prefix + k: zero_like(v) for k, v in tensor_pack.items()}
    elif isinstance(tensor_pack, torch.Tensor):
        return tensor_pack.new_zeros(tensor_pack.shape)
    elif isinstance(tensor_pack, np.ndarray):
        return np.zeros_like(tensor_pack)
    else:
        warnings.warn("Unexpected data type {}".format(type(tensor_pack)))
        return 0


def pad_stack(tensors, shape, pad_value=255):
    tensors = torch.stack(
        [
            F.pad(
                tensor,
                pad=[0, shape[1] - tensor.shape[1], 0, shape[0] - tensor.shape[0]],
                value=pad_value,
            )
            for tensor in tensors
        ]
    )
    return tensors


def result2bbox(result):
    num_class = len(result)

    bbox = np.concatenate(result)
    if bbox.shape[0] == 0:
        label = np.zeros(0, dtype=np.uint8)
    else:
        label = np.concatenate(
            [[i] * len(result[i]) for i in range(num_class) if len(result[i]) > 0]
        ).reshape((-1,))
    return bbox, label


def result2mask(result):
    num_class = len(result)
    mask = [np.stack(result[i]) for i in range(num_class) if len(result[i]) > 0]
    if len(mask) > 0:
        mask = np.concatenate(mask)
    else:
        mask = np.zeros((0, 1, 1))
    return BitmapMasks(mask, mask.shape[1], mask.shape[2]), None


def sequence_mul(obj, multiplier):
    if isinstance(obj, Sequence):
        return [o * multiplier for o in obj]
    else:
        return obj * multiplier


def is_match(word, word_list):
    for keyword in word_list:
        if keyword in word:
            return True
    return False


def weighted_loss(loss: dict, weight, ignore_keys=[], warmup=0):
    _step_counter["weight"] += 1
    lambda_weight = (
        lambda x: x * (_step_counter["weight"] - 1) / warmup
        if _step_counter["weight"] <= warmup
        else x
    )
    if isinstance(weight, Mapping):
        for k, v in weight.items():
            for name, loss_item in loss.items():
                if (k in name) and ("loss" in name):
                    loss[name] = sequence_mul(loss[name], lambda_weight(v))
    elif isinstance(weight, Number):
        for name, loss_item in loss.items():
            if "loss" in name:
                if not is_match(name, ignore_keys):
                    loss[name] = sequence_mul(loss[name], lambda_weight(weight))
                else:
                    loss[name] = sequence_mul(loss[name], 0.0)
    else:
        raise NotImplementedError()
    return loss



from fileinput import filename
import sys
import os
import json
import logging
from antgo.utils import args
system_path = os.path.join(os.path.abspath(os.curdir),'system.py')
os.system(f'ln -sf {system_path}  {os.path.dirname(os.path.realpath(__file__))}/system.py')
from antgo.framework.helper.fileio import *
from antgo.framework.helper.utils import Config
from system import *


args.DEFINE_nn_args()

def main():
    nn_args = args.parse_args()
    # 下载扩展配置中的数据到本地
    if os.path.exists(nn_args.extra_config):
        with open(nn_args.extra_config, 'r') as fp:
            extra_config = Config.fromstring(fp.read(),'.json')

        # extra_config 格式为项目信息格式
        # step3.1: 数据相关下载 (默认TFRECORD是antgo默认标准打包格式)
        extra_dataset_train_label = extra_config['source'].get('label', [])
        extra_dataset_train_pseudo_label = extra_config['source'].get('pseudo-label', [])
        extra_dataset_train_unlabel = extra_config['source'].get('unlabel', [])

        if not os.path.exists('dataset-storage'):
            os.mkdir('dataset-storage')

        # 有标签数据下载，伪标签数据下载，无标签数据下载
        for data_stage, data_info_list in zip(
            ['label', 'pseudo-label', 'unlabel'], 
            [extra_dataset_train_label, extra_dataset_train_pseudo_label, extra_dataset_train_unlabel]):

            # 下载相关数据，到训练集群
            if not os.path.exists(f"dataset-storage/{data_stage}/"):
                os.makedirs(f"dataset-storage/{data_stage}/")

            for data_info in data_info_list:
                if data_info['status'] and data_info['address'] != '':
                    if data_info['address'].startswith('/'):
                        # 本地路径
                        data_folder = os.path.dirname(data_info['address'])
                        target_data_folder = f"dataset-storage/{data_stage}"

                        check_prefix = data_info['address'].split('/')[-1]
                        for file_name in os.listdir(data_folder):
                            if os.path.exists(f'{target_data_folder}/{file_name}'):
                                # 目标文件已经存在，直接跳过
                                continue

                            if file_name.startswith(check_prefix):
                                os.system(f'ln -s {data_folder}/{file_name} {target_data_folder}/{file_name}')                    
                    else:
                        check_prefix = data_info['address'].split('/')[-1]
                        is_existed = False
                        for file_name in os.listdir(f"dataset-storage/{data_stage}"):
                            if file_name.startswith(check_prefix):
                                is_existed = True
                                break

                        if is_existed:
                            continue

                        if not data_info['address'].endswith('*'):
                            data_info['address'] += '*'
                        status = file_client_get(data_info['address'], f"dataset-storage/{data_stage}")
                        if not status:
                            logging.error(f'Download {data_info["address"]} error.')
                    
    # 下载配置中的数据到本地
    if not os.path.exists(nn_args.config):
        # 查找位置1
        config_file_path = os.path.join(nn_args.exp.split('.')[0], 'config.py')
        if not os.path.exists(config_file_path):
            config_file_path = ''

        # 查找位置2
        if config_file_path == '':
            config_file_path = os.path.join(nn_args.exp.split('.')[0], 'configs', 'config.py')
            if not os.path.exists(config_file_path):
                config_file_path = ''
        nn_args.config = config_file_path

    if os.path.exists(nn_args.config):
        if not os.path.exists('dataset-storage'):
            os.mkdir('dataset-storage')

        cfg = Config.fromfile(nn_args.config)
        for data_stage in ['train', 'val', 'test']:
            if not os.path.exists(f"dataset-storage/{data_stage}/"):
                os.makedirs(f"dataset-storage/{data_stage}/")

            for data_record_path in getattr(cfg.data, data_stage).data_path_list:
                if data_record_path.startswith('/') or data_record_path.startswith('./'):
                    # 本地路径
                    data_folder = os.path.dirname(data_record_path)
                    target_data_folder = f"dataset-storage/{data_stage}"

                    check_prefix = data_record_path.split('/')[-1]
                    for file_name in os.listdir(data_folder):
                        if os.path.exists(f'{target_data_folder}/{file_name}'):
                            # 目标文件已经存在，直接跳过
                            continue

                        if file_name.startswith(check_prefix):
                            os.system(f'ln -s {data_folder}/{file_name} {target_data_folder}/{file_name}')                    
                else:
                    check_prefix = data_record_path.split('/')[-1]
                    is_existed = False
                    for file_name in os.listdir(f"dataset-storage/{data_stage}"):
                        if file_name.startswith(check_prefix):
                            is_existed = True
                            break

                    if is_existed:
                        continue

                    if not data_record_path.endswith('*'):
                        data_record_path += '*'
                    status = file_client_get(data_record_path, f"dataset-storage/{data_stage}")
                    if not status:
                        logging.error(f'Download {data_record_path} error.')

    # 下载checkpoint到本地
    if nn_args.checkpoint is not None and nn_args.checkpoint != '':
        if not os.path.exists('checkpoint-storage'):
            os.mkdir('checkpoint-storage')

        if not nn_args.checkpoint.startswith('/') and not nn_args.checkpoint.startswith('./'):
            status = file_client_get(nn_args.checkpoint, f"checkpoint-storage")
            if not status:
                logging.error(f'Download checkpoint fail.')

    if nn_args.resume_from is not None and nn_args.resume_from != '':
        if not os.path.exists('checkpoint-storage'):
            os.mkdir('checkpoint-storage')

        if not nn_args.checkpoint.startswith('/') and not  nn_args.checkpoint.startswith('./'):
            status = file_client_get(nn_args.resume_from, f"checkpoint-storage")
            if not status:
                logging.error(f'Download checkpoint fail.')

    
if __name__ == "__main__":
    main()


import logging
import os
from queue import PriorityQueue
import shutil
import yaml 
import json
from antgo import config
from antgo.script.base import *


# 提交任务运行
def ssh_submit_process_func(project_name, sys_argv, gpu_num, cpu_num, memory_size, task_name=None):   
    # 前提假设，调用此函数前当前目录下需要存在项目代码
    # step1: 加载ssh配置文件
    ssh_submit_config_file = os.path.join(os.environ['HOME'], '.config', 'antgo', 'ssh-submit-config.yaml')
    if not os.path.exists(ssh_submit_config_file):
        logging.error('No ssh submit config.')
        return False

    with open(ssh_submit_config_file, encoding='utf-8', mode='r') as fp:
        config_content = yaml.safe_load(fp)

    username = config_content['config']['username']
    password = config_content['config']['password']
    ip = config_content['config']['ip']
    submit_script = os.path.join(os.path.dirname(__file__), 'ssh-submit.sh')

    with open(os.path.join(config.AntConfig.task_factory,f'{project_name}.json'), 'r') as fp:
        project_info = json.load(fp)

    image_name = '' # 基础镜像
    if project_info['image'] != '':
        image_name = project_info['image']
    if image_name == '':
        image_name = 'antgo-env:latest'
    
    # 添加扩展配置：保存到当前目录下并一同提交
    if task_name is not None:
        extra_config = prepare_extra_config(task_name, project_info)
        if extra_config is None:
            return False

        with open('./extra-config.py', 'w') as fp:
            json.dump(extra_config, fp)    
        sys_argv += " --extra-config=./extra-config.py"

    # 执行提交命令
    if password == '':
        password = 'default'
    submit_cmd = f'bash {submit_script} {username} {password} {ip} {gpu_num} {cpu_num} {memory_size}M "{sys_argv}" {image_name} {project_name}'
    os.system(submit_cmd)

    # 删除临时配置：
    if os.path.exists('./extra-config.py'):
        os.remove('./extra-config.py')
    return True

# 检查任务资源是否满足
def ssh_submit_resource_check_func(gpu_num, cpu_num, memory_size):
    # TODO，支持资源检查
    return True


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from ...builder import DISTILL_LOSSES

'''
Modified from https://github.com/clovaai/overhaul-distillation/blob/master/CIFAR-100/distiller.py
'''


@DISTILL_LOSSES.register_module()
class OFD(nn.Module):
	'''
	A Comprehensive Overhaul of Feature Distillation
	http://openaccess.thecvf.com/content_ICCV_2019/papers/
	Heo_A_Comprehensive_Overhaul_of_Feature_Distillation_ICCV_2019_paper.pdf
	'''
	def __init__(self, in_channels, out_channels):
		super(OFD, self).__init__()
		self.connector = nn.Sequential(*[
				nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=1, padding=0, bias=False),
				nn.BatchNorm2d(out_channels)
			])

		for m in self.modules():
			if isinstance(m, nn.Conv2d):
				nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
				if m.bias is not None:
					nn.init.constant_(m.bias, 0)
			elif isinstance(m, nn.BatchNorm2d):
				nn.init.constant_(m.weight, 1)
				nn.init.constant_(m.bias, 0)

	def forward(self, fm_s, fm_t):
		margin = self.get_margin(fm_t)
		fm_t = torch.max(fm_t, margin)
		fm_s = self.connector(fm_s)

		mask = 1.0 - ((fm_s <= fm_t) & (fm_t <= 0.0)).float()
		loss = torch.mean((fm_s - fm_t)**2 * mask)

		return loss

	def get_margin(self, fm, eps=1e-6):
		mask = (fm < 0.0).float()
		masked_fm = fm * mask

		margin = masked_fm.sum(dim=(0,2,3), keepdim=True) / (mask.sum(dim=(0,2,3), keepdim=True)+eps)

		return margin


# Copyright (c) 2021  PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import paddle
from PIL import Image
import os
import os.path
from typing import Any, Callable, cast, Dict, List, Optional, Tuple
IMG_EXTENSIONS = ('.jpg', '.jpeg', '.png', '.ppm', '.bmp', '.pgm', '.tif',
                  '.tiff', '.webp')


def has_file_allowed_extension(filename: str,
                               extensions: Tuple[str, ...]) -> bool:
    return filename.lower().endswith(extensions)


def make_dataset(
        directory: str,
        class_to_idx: Dict[str, int],
        extensions: Optional[Tuple[str, ...]]=None,
        is_valid_file: Optional[Callable[[str], bool]]=None, ) -> List[Tuple[
            str, int]]:
    instances = []
    directory = os.path.expanduser(directory)
    both_none = extensions is None and is_valid_file is None
    both_something = extensions is not None and is_valid_file is not None
    if both_none or both_something:
        raise ValueError(
            "Both extensions and is_valid_file cannot be None or not None at the same time"
        )
    if extensions is not None:

        def is_valid_file(x: str) -> bool:
            return has_file_allowed_extension(x,
                                              cast(Tuple[str, ...], extensions))

    is_valid_file = cast(Callable[[str], bool], is_valid_file)
    for target_class in sorted(class_to_idx.keys()):
        class_index = class_to_idx[target_class]
        target_dir = os.path.join(directory, target_class)
        if not os.path.isdir(target_dir):
            continue
        for root, _, fnames in sorted(os.walk(target_dir, followlinks=True)):
            for fname in sorted(fnames):
                path = os.path.join(root, fname)
                if is_valid_file(path):
                    item = path, class_index
                    instances.append(item)
    return instances


class ImageFolder(paddle.vision.datasets.ImageFolder):
    def __init__(self,
                 root,
                 transform=None,
                 target_transform=None,
                 loader=None,
                 is_valid_file=None):
        super().__init__(
            root,
            loader=loader,
            transform=transform,
            is_valid_file=is_valid_file)
        self.target_transform = target_transform
        classes, class_to_idx = self._find_classes(root)
        self.samples = self.make_dataset(self.root, class_to_idx, IMG_EXTENSIONS
                                         if is_valid_file is None else None,
                                         is_valid_file)

    def _find_classes(self, dir: str) -> Tuple[List[str], Dict[str, int]]:
        classes = [d.name for d in os.scandir(dir) if d.is_dir()]
        classes.sort()
        class_to_idx = {cls_name: i for i, cls_name in enumerate(classes)}
        return classes, class_to_idx

    @staticmethod
    def make_dataset(
            directory: str,
            class_to_idx: Dict[str, int],
            extensions: Optional[Tuple[str, ...]]=None,
            is_valid_file: Optional[Callable[[str], bool]]=None, ) -> List[
                Tuple[str, int]]:
        return make_dataset(
            directory,
            class_to_idx,
            extensions=extensions,
            is_valid_file=is_valid_file)

    def __getitem__(self, index):
        path, target = self.samples[index]
        sample = self.loader(path)
        if self.transform is not None:
            sample = self.transform(sample)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return [sample, target]



import torch.nn as nn
import torch
import torch.nn.functional as F
from antgo.framework.helper.models.builder import BACKBONES
from antgo.framework.helper.cnn import ConvModule,DepthwiseSeparableConvModule

def make_divisible(v, divisor=8, min_value=None):
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v

class StemConv(nn.Module):
    def __init__(self, in_channels, out_channels, kernel_size, stride, num_groups=1):
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding=(kernel_size - 1)//2,
            groups=num_groups,
            bias=False)
        self.bn = nn.BatchNorm2d(out_channels)
        # self.hardswish = nn.Hardswish()
        self.hardswish = nn.ReLU()

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.hardswish(x)
        return x

class SEModule(nn.Module):
    def __init__(self, channel, reduction=4):
        super().__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.conv1 = nn.Conv2d(
            in_channels=channel,
            out_channels=channel // reduction,
            kernel_size=1,
            stride=1,
            padding=0)
        self.relu = nn.ReLU()
        self.conv2 = nn.Conv2d(
            in_channels=channel // reduction,
            out_channels=channel,
            kernel_size=1,
            stride=1,
            padding=0)
        self.hardsigmoid = nn.Hardsigmoid()

    def forward(self, x):
        identity = x
        x = self.avg_pool(x)
        x = self.conv1(x)
        x = self.relu(x)
        x = self.conv2(x)
        x = self.hardsigmoid(x)
        out = identity * x
        return out

class DepthSepConvBlock(nn.Module):
    def __init__(self, in_channels, out_channels, stride, dw_kernel_size, use_se=False, pw_kernel_size=1):
        super().__init__()
        self.use_se = use_se
        self.dw_conv = StemConv(
            in_channels, in_channels, kernel_size=dw_kernel_size,
            stride=stride, num_groups=in_channels)
        if self.use_se:
            self.se = SEModule(in_channels)
        self.pw_conv = StemConv(in_channels, out_channels, kernel_size=pw_kernel_size, stride=1)

    def forward(self, x):
        x = self.dw_conv(x)
        if self.use_se:
            x = self.se(x)
        x = self.pw_conv(x)

        return x

class DPM(nn.Module):
    def __init__(self, in_channels, middle_channels=32, out_channels=32):
        super().__init__()
        self.multi_block_0 = ConvModule(
                    in_channels, 
                    middle_channels,
                    kernel_size=1,
                    padding=0, 
                    act_cfg=dict(type='ReLU'), 
                    norm_cfg=dict(type='BN')
        )        
        self.multi_block_1 = DepthwiseSeparableConvModule(
            in_channels=middle_channels,
            out_channels=middle_channels,
            kernel_size=3,
            padding=(int)((2*(3-1)+1-1)//2),
            dilation=2,
            dw_act_cfg=dict(type='ReLU'), dw_norm_cfg=dict(type='BN'), pw_norm_cfg=dict(type='BN'),pw_act_cfg=dict(type='ReLU')   
        )
        self.multi_block_2 = DepthwiseSeparableConvModule(
            in_channels=middle_channels,
            out_channels=middle_channels,
            kernel_size=3,
            padding=(int)((4*(3-1)+1-1)//2),
            dilation=4,
            dw_act_cfg=dict(type='ReLU'), dw_norm_cfg=dict(type='BN'), pw_norm_cfg=dict(type='BN'),pw_act_cfg=dict(type='ReLU')   
        )
        self.multi_block_fusion = ConvModule(
                middle_channels*3,
                out_channels,
                1,
                padding=0,
                act_cfg=dict(type='ReLU'), 
                norm_cfg=dict(type='BN'),
        )        
    
    def forward(self, x):
        x0 = self.multi_block_0(x)
        x1 = self.multi_block_1(x0)
        x2 = self.multi_block_2(x1)
        x3 = torch.cat([x0,x1,x2], dim=1)
        x = self.multi_block_fusion(x3)
        return x


@BACKBONES.register_module()
class DDRLCNet(nn.Module):
    def __init__(self, cfgs, in_channels=1, block=DepthSepConvBlock, num_classes=1000, dropout=0.2, scale=1.0, class_expand=1280):
        super(DDRLCNet, self).__init__()
        self.cfgs = cfgs
        self.class_expand = class_expand
        self.block = block

        self.conv1 = StemConv(
            in_channels=in_channels,
            kernel_size=3,
            out_channels=make_divisible(16 * scale),
            stride=2)
        self.stages = torch.nn.ModuleList()
        self.channels_num = {}
        for cfg_i, cfg in enumerate(self.cfgs):
            layers = []
            out_channel = 0
            for k, inplanes, planes, stride, use_se in cfg:
                in_channel = make_divisible(inplanes * scale)
                out_channel = make_divisible(planes * scale)
                layers.append(block(in_channel, out_channel, stride=stride, dw_kernel_size=k, use_se=use_se))

            self.channels_num[cfg_i] = out_channel
            self.stages.append(nn.Sequential(*layers))

        # 2   ,  4,   10,   12
        # 1/4 ,  1/8, 1/16, 1/32
        # (dilation(k-1)+1-stride)/2
        self.down4_to_16 = \
                ConvModule(
                    32, 
                    self.channels_num[10],
                    kernel_size=3,
                    stride=4, 
                    padding=(int)((4*(3-1)+1-4)//2+1), 
                    dilation=4, 
                    act_cfg=None, 
                    norm_cfg=dict(type='BN'))
        self.down4_to_32 = \
                ConvModule(
                    32, 
                    self.channels_num[12],
                    kernel_size=3,
                    stride=8, 
                    padding=(int)((6*(3-1)+1-8)//2+1), 
                    dilation=6, 
                    act_cfg=None, 
                    norm_cfg=dict(type='BN'))             

        self.compression8_to_4 = \
                ConvModule(
                    self.channels_num[4], 
                    32, 
                    kernel_size=1, 
                    act_cfg=None, 
                    norm_cfg=dict(type='BN'))
        self.compression16_to_4 = \
                ConvModule(
                    self.channels_num[10], 
                    32, 
                    kernel_size=1, 
                    act_cfg=None, 
                    norm_cfg=dict(type='BN'))

        self.layer4_1 = DepthwiseSeparableConvModule(
            in_channels=self.channels_num[2],
            out_channels=32,
            kernel_size=3,
            padding=1,
            dilation=1,
            act_cfg=None, dw_act_cfg=dict(type='ReLU'), dw_norm_cfg=dict(type='BN'), pw_norm_cfg=dict(type='BN')
        )
        self.layer4_2 = DepthwiseSeparableConvModule(
            in_channels=32,
            out_channels=32,
            kernel_size=3,
            padding=1,
            dilation=1,
            act_cfg=None, dw_act_cfg=dict(type='ReLU'), dw_norm_cfg=dict(type='BN'), pw_norm_cfg=dict(type='BN')
        )
        self.layer4_3 = DepthwiseSeparableConvModule(
            in_channels=32,
            out_channels=32,
            kernel_size=3,
            padding=1,
            dilation=1,
            act_cfg=None, dw_act_cfg=dict(type='ReLU'), dw_norm_cfg=dict(type='BN'), pw_norm_cfg=dict(type='BN')
        )

        # (dilation(k-1)+1-stride)/2
        self.multi_block_0 = DepthwiseSeparableConvModule(
            in_channels=32,
            out_channels=32,
            kernel_size=3,
            padding=(int)((2*(3-1)+1-1)//2),
            dilation=2,
            dw_act_cfg=dict(type='ReLU'), dw_norm_cfg=dict(type='BN'), pw_norm_cfg=dict(type='BN'),pw_act_cfg=dict(type='ReLU')     
        )
        self.multi_block_1 = DepthwiseSeparableConvModule(
            in_channels=32,
            out_channels=32,
            kernel_size=3,
            padding=(int)((4*(3-1)+1-1)//2),
            dilation=4,
            dw_act_cfg=dict(type='ReLU'), dw_norm_cfg=dict(type='BN'), pw_norm_cfg=dict(type='BN'),pw_act_cfg=dict(type='ReLU')   
        )
        self.multi_block_2 = DepthwiseSeparableConvModule(
            in_channels=32,
            out_channels=32,
            kernel_size=3,
            padding=(int)((6*(3-1)+1-1)//2),
            dilation=6,
            dw_act_cfg=dict(type='ReLU'), dw_norm_cfg=dict(type='BN'), pw_norm_cfg=dict(type='BN'),pw_act_cfg=dict(type='ReLU')   
        )
        self.multi_block_fusion = ConvModule(
                128,
                32,
                1,
                padding=0,
                norm_cfg=dict(type='BN'),
        )        

        # DPM
        self.dpm = DPM(self.channels_num[12], 32, 32)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

    def forward(self, x):
        x = self.conv1(x)
        out_layer = None # 1/4 resolution
        for stage_i, stage_module in enumerate(self.stages):
            x = stage_module(x)
            if stage_i == 2:
                out_layer = self.layer4_1(x)        # channel 32
            elif stage_i == 4:
                # 1/8
                # merge from 1/4
                x_ = self.layer4_2(out_layer)       # channel 32
                out_layer = x_ + F.interpolate(
                                self.compression8_to_4(x),
                                scale_factor=2,
                                mode='bilinear')
                out_layer = F.relu(out_layer)       # 1/4
            elif stage_i == 10:
                # 1/16
                # merge from 1/4
                down_x = F.relu(x + self.down4_to_16(out_layer))
                x_ = self.layer4_3(out_layer)
                out_layer = x_ + F.interpolate(
                                self.compression16_to_4(x),
                                scale_factor=4,
                                mode='bilinear')
                out_layer = F.relu(out_layer)
                x = down_x     
            elif stage_i == 12:
                # 1/32
                down_x = F.relu(x + self.down4_to_32(out_layer))
                # x_ = self.layer4_4(out_layer)
                # out_layer = x_ + F.interpolate(
                #                 self.compression32_to_4(x),
                #                 scale_factor=8,
                #                 mode='bilinear')    
                x = down_x

        # fuse 1/4 and 1/32
        out_layer = out_layer + F.interpolate(self.dpm(x), scale_factor=8, mode='bilinear') 

        # 瀑布式结构（多dilation操作实现视野捕捉）
        branch_1 = self.multi_block_0(out_layer)
        branch_2 = self.multi_block_1(branch_1)
        branch_3 = self.multi_block_2(branch_2)
        branch_all = torch.cat([out_layer, branch_1, branch_2, branch_3], dim=1)
        out_layer = self.multi_block_fusion(branch_all)
        return out_layer,
    
    def init_weights(self):
        pass


import sys
from fileinput import filename
from random import shuffle
import shutil
import cv2
import os
import json
import copy
import glob
import logging
import numpy as np
from pprint import pprint
import imagesize
import requests
from antgo.utils.sample_gt import *
from pycocotools.coco import COCO


def extract_from_videos(video_folder, target_folder, frame_rate=10, max_size=0, **kwargs):
    # 输出
    # -data
    # -annotation.json
    if target_folder is None:
        target_folder = './'
    
    if not os.path.exists(target_folder):
        os.makedirs(target_folder)

    # 不支持多级目录
    # 抽帧过程
    support_video_ext = ['mp4', 'avi']
    frame_height = 0
    frame_width = 0
    for video_file_name in os.listdir(video_folder):
        if video_file_name[0] == '.':
            continue
        if video_file_name.split('.')[-1].lower() not in support_video_ext:
            continue
        
        video_file_pure_name = video_file_name.split('.')[0]
        if not os.path.exists(os.path.join(target_folder, video_file_pure_name)):
            os.makedirs(os.path.join(target_folder, video_file_pure_name))
        video_file_path = os.path.join(video_folder, video_file_name)
        cap = cv2.VideoCapture(video_file_path) 
        print(f'process video {video_file_name}')
        count = 0
        while(cap.isOpened()): 
            ret, frame = cap.read() 
            if not ret:
                break
            
            frame_height, frame_width = frame.shape[:2]
            
            if max_size > 0:
                scale = 1.0
                if frame_height > frame_width:
                    scale = max_size / frame_height
                else:
                    scale = max_size / frame_width
                frame = cv2.resize(frame, None, None, fx=scale, fy=scale)
                
            if count % frame_rate == 0:
                cv2.imwrite(os.path.join(target_folder, video_file_pure_name, f'{video_file_pure_name}_rate-{frame_rate}_frame-{count}.png'), frame)

            count += 1

        cap.release() 

    # 生成默认格式解析json
    annotation_file_name = 'annotation.json'
    annotation_list = []
    # 加载默认标准格式
    sample_gt_file = os.path.join('/'.join(os.path.dirname(__file__).split('/')[:-1]), 'resource', 'templates', 'sample_gt.json')
    with open(sample_gt_file, 'r', encoding="utf-8") as fp:
        sample_gt = json.load(fp)

    for video_name in os.listdir(target_folder):
        if video_name[0] == '.':
            continue
        if not os.path.isdir(os.path.join(target_folder, video_name)):
            continue

        for file_name in os.listdir(os.path.join(target_folder, video_name)):
            if file_name[0] == '.':
                continue
            
            sample_gt_cp = copy.deepcopy(sample_gt)
            sample_gt_cp.update({
                'image_file': f'{video_name}/{file_name}'
            })
            sample_gt_cp['height'] = frame_height
            sample_gt_cp['width'] = frame_width
            annotation_list.append(sample_gt_cp)

    print(f'Extract frame number {len(annotation_list)}')

    with open(os.path.join(target_folder, annotation_file_name), 'w', encoding="utf-8") as fp:
        json.dump(annotation_list, fp)
    

def extract_from_images(
    source_folder, 
    target_folder, 
    filter_prefix=None, 
    filter_suffix=None, 
    filter_ext=None, 
    shuffle=False, num=0, max_size=0, **kwargs):
    support_image_ext = ['png', 'jpeg', 'jpg']
    if filter_ext is not None:
        temp = []
        for ext in support_image_ext:
            if ext in filter_ext:
                temp.append(ext)
        
        support_image_ext = temp

    if target_folder is None:
        target_folder = './'

    if not os.path.exists(target_folder):
        os.makedirs(target_folder)
        
    # 遍历source_folder 所有文件并保留文件目录结构复制
    sample_file_list = []
    for root, dirnames, filenames in os.walk(source_folder):
        for filename in filenames:
            if filename[0] == '.':
                continue
            if filename.split('/')[-1].split('.')[-1].lower() not in support_image_ext:
                continue
            if filter_prefix is not None:
                if not filename.startswith(filter_prefix):
                    continue
            
            if filter_suffix is not None:
                if not filename.split('.')[0].endswith(filter_suffix):
                    continue
            
            if filter_ext is not None:
                if not filename.endswith(filter_ext):
                    continue
                    
            # copy to 对应目录
            source_path = os.path.join(root, filename)
            rel_path = os.path.relpath(source_path, source_folder)
            sample_file_list.append(rel_path)

    if shuffle:
        np.random.shuffle(sample_file_list)
    
    if num <= 0:
        num = len(sample_file_list)
    
    # 生成默认格式解析json
    annotation_file_name = 'annotation.json'
    annotation_list = []
    # 加载默认标准格式
    sample_gt_file = os.path.join('/'.join(os.path.dirname(__file__).split('/')[:-1]), 'resource', 'templates', 'sample_gt.json')
    with open(sample_gt_file, 'r', encoding="utf-8") as fp:
        sample_gt = json.load(fp)

    for sample_file_name in sample_file_list[:num]:
        source_path = os.path.join(source_folder, sample_file_name)
        target_path = os.path.join(target_folder, sample_file_name)
        dir_path = os.path.dirname(target_path)
        if not os.path.exists(os.path.join(dir_path)):
            os.makedirs(dir_path)
        
        if max_size > 0:
            image = cv2.imread(source_path)
            height, width = image.shape[:2]
            scale = 1.0
            if height > width:
                scale = max_size / height
            else:
                scale = max_size / width
            image = cv2.resize(image, None, None, fx=scale, fy=scale)
            cv2.imwrite(target_path, image)
        else:  
            shutil.copy(source_path, target_path)
            
        sample_file_path = os.path.join(target_folder, sample_file_name)
        width, height = imagesize.get(sample_file_path)
    
        sample_gt_cp = copy.deepcopy(sample_gt)
        sample_gt_cp.update({
            'image_file': sample_file_name
        })
        sample_gt_cp['height'] = height
        sample_gt_cp['width'] = width
        annotation_list.append(sample_gt_cp)

    print(f'extract image number {len(annotation_list)}')

    with open(os.path.join(target_folder, annotation_file_name), 'w', encoding="utf-8") as fp:
        json.dump(annotation_list, fp)


def extract_from_coco(source_file, target_folder, filter_label=None, **kwargs):
    if not source_file.endswith('json'):
        logging.error('Source file must be json file')
        return

    if filter_label is None:
        filter_label = ''    
    filter_label_map = {}
    for ll in filter_label.split(','):
        label_name, label_index = ll.split(":")
        filter_label_map[label_name] = int(label_index)
    
    if target_folder is None:
        target_folder = './'
    
    if not os.path.exists(target_folder):
        os.makedirs(target_folder)
    
    total_samples = []
    sgt = SampleGTTemplate() 
    db = COCO(source_file)
    for aid in db.anns.keys():
        ann = db.anns[aid]
        image_id = ann['image_id']
        img_info = db.loadImgs(image_id)[0]
        file_name = img_info['file_name']
        width = img_info['width']
        height = img_info['height']
        
        gt_template = sgt.get()
        gt_template['image_file'] = file_name
        gt_template['width'] = width
        gt_template['height'] = height
        anns = db.imgToAnns[image_id]
        
        is_ignore = True
        for ins_info in anns:
            ins_segmentation = ins_info['segmentation']
            ins_bbox = ins_info['bbox']
            box_x, box_y, box_w, box_h = ins_bbox
            ins_category = ins_info['category_id']
            ins_category_name = db.cats[ins_category]['name']
            
            if len(filter_label_map) > 0:
                if ins_category_name not in filter_label_map:
                    continue
                ins_category = filter_label_map[ins_category_name]
                
            is_ignore = False
            if len(ins_bbox) > 0:
                gt_template['labels'].append(ins_category)
                gt_template['label_names'].append(ins_category_name)
                gt_template['bboxes'].append([box_x, box_y, box_x+box_w, box_y+box_h])
            
                gt_template['segments'].append(ins_segmentation)
                if len(ins_segmentation) > 0:
                    gt_template['has_segments'].append(1)
                else:
                    gt_template['has_segments'].append(0)

        if not is_ignore:
            total_samples.append(gt_template)
        
    with open(os.path.join(target_folder, 'annotation.json'), 'w', encoding="utf-8") as fp:
        json.dump(total_samples, fp)

    with open(os.path.join(target_folder, 'meta.json'), 'w', encoding="utf-8") as fp:
        json.dump(sgt.meta(), fp)


def extract_from_crop(source_file, target_folder, **kwargs):
    # 仅支持标准GT
    source_file_name = source_file.split('/')[-1]
    pure_file_name = source_file_name.split('.')[0]
    ext_name =source_file_name.split('.')[-1]
    if ext_name != 'json':
        logging.error('Only support json or txt file.')
        return
    
    if target_folder is None:
        target_folder = './'

    if not os.path.exists(target_folder):
        os.makedirs(target_folder)

    if not os.path.exists(os.path.join(target_folder, 'images')):
        os.makedirs(os.path.join(target_folder, 'images'))

    with open(source_file, 'r', encoding="utf-8") as fp:
        content = json.load(fp)

    src_folder = os.path.dirname(source_file)
    sgtt = SampleGTTemplate()
    total_gt_list = []
    for sample_i, sample in enumerate(content):
        bboxes = sample['bboxes']
        labels = sample['labels']
        label_names = sample['label_names']
        if len(label_names) == 0:
            label_names = [str(l) for l in labels]
        
        image = None
        file_name = f'{str(sample_i)}.png'
        if sample['image_file'] != '':
            image_path = os.path.join(src_folder, sample['image_file'])
            try:
                image = cv2.imread(image_path)
                file_name = image_path.split('/')[-1]
            except:
                logging.error("Couldnt imread %s."%sample['image_file'])
                image = None
        else:
            try:
                pic = requests.get(sample['image_url'], timeout=20)
                image = cv2.imdecode(np.frombuffer(pic.content, np.uint8), cv2.IMREAD_COLOR)
            except:
                logging.error("Couldnt download %s."%sample['image_url'])
                image = None

        if image is None:
            continue
        
        ext_name = file_name.split('.')[-1]
        pure_file_name = file_name.split(f'.{ext_name}')[0]
        height, width = image.shape[:2]
        
        for box_i, (box, label, label_name) in enumerate(zip(bboxes, labels, label_names)):
            x0,y0,x1,y1 = box
            
            box_w = x1 - x0
            box_h = y1 - y0
            
            if kwargs.get('ext_ratio') > 0:
                box_w = box_w * (1+kwargs['ext_ratio'])
                box_h = box_h * (1+kwargs['ext_ratio'])
            box_size = max(box_w, box_h)
            
            cx = (x0+x1)/2.0
            cy = (y0+y1)/2.0
               
            x0 = int(max(cx-box_size/2, 0))
            y0 = int(max(cy-box_size/2, 0))
            x1 = int(min(cx+box_size/2, width))
            y1 = int(min(cy+box_size/2, height))
            
            if x0 == x1 or y0 == y1:
                continue
                        
            patch_image = image[y0:y1,x0:x1]
            patch_file_name = f'{pure_file_name}_box_{box_i}_label_{label_name}.png'
            subfolder = label_name
            if not os.path.exists(os.path.join(target_folder, 'images', subfolder)):
                os.makedirs(os.path.join(target_folder, 'images', subfolder))
            cv2.imwrite(os.path.join(target_folder, 'images', subfolder, patch_file_name), patch_image)
            
            standard_gt = sgtt.get()
            standard_gt['image_file'] = f'images/{subfolder}/{patch_file_name}'
            standard_gt['height'] = patch_image.shape[0]
            standard_gt['width'] = patch_image.shape[1]
            standard_gt['image_label'] = label
            standard_gt['image_label_name'] = label_name
            
            if len(sample['has_joints2d']) > 0:
                if sample['has_joints2d'][box_i]:
                    standard_gt['has_joints2d'].append(1)
                    standard_gt['joints2d'].append(sample['joints2d'][box_i])
                else:
                    standard_gt['has_joints2d'].append(0)
                    standard_gt['joints2d'].append([])
            
            if len(sample['has_segments']) > 0:
                if sample['has_segments'][box_i]:
                    standard_gt['has_segments'].append(1)
                    standard_gt['segments'].append(sample['segments'][box_i])
                else:
                    standard_gt['has_segments'].append(0)
                    standard_gt['segments'].append([])
            
            total_gt_list.append(standard_gt)

    with open(os.path.join(target_folder, 'annotation.json'), 'w', encoding="utf-8") as fp:
        json.dump(total_gt_list, fp)

    with open(os.path.join(target_folder, 'meta.json'), 'w', encoding="utf-8") as fp:
        json.dump(sgtt.meta(), fp)


def extract_from_samples(source_file, target_folder, num=1, feedback=False, **kwargs):
    source_file_name = source_file.split('/')[-1]
    pure_file_name = source_file_name.split('.')[0]
    ext_name =source_file_name.split('.')[-1]
    if ext_name not in ['json', 'txt']:
        logging.error('Only support json or txt file.')
        return
    
    if target_folder is None:
        target_folder = './'
    
    # 随机抽取10条数据
    samples = []
    sample_indexs = []
    total_sample_num = 0
    if ext_name == 'json':
        with open(source_file, 'r', encoding="utf-8") as fp:
            content = json.load(fp)
            total_sample_num = len(content)
            sample_num = min(total_sample_num, num)

            sample_indexs = np.random.choice(total_sample_num, sample_num, replace=False).tolist()
            for i in sample_indexs:
                samples.append(content[i])

        target_file_path = os.path.join(target_folder, f'{pure_file_name}_samples_{len(samples)}.{ext_name}')
        with open(target_file_path, 'w', encoding="utf-8") as fp:
            json.dump(samples, fp)

        if feedback:
            # step1: 显示样本总条数
            print(f'Total Sample Num: {total_sample_num}')
            
            # step2: 显示keys
            if isinstance(samples[0], dict):
                print('Total Keys: ')
                print(samples[0].keys())

            # step3: 显示采样的索引
            print('Sample Index: ')
            print(sample_indexs)

            # step4: 显示采样的样本
            print('Samples: ')
            pprint(samples)
    else:
        with open(source_file, 'r', encoding="utf-8") as fp:
            content = fp.readlines()

            total_sample_num = len(content)
            sample_num = min(total_sample_num, num)
            sample_indexs = np.random.choice(total_sample_num, sample_num, replace=False).tolist()
            for i in sample_indexs:
                samples.append(content[i].strip())
        
        target_file_path = os.path.join(target_folder, f'{pure_file_name}_samples_{len(samples)}.{ext_name}')
        with open(target_file_path, 'w', encoding="utf-8") as fp:
            for s in samples:
                fp.write(f'{s}\n')

        if feedback:
            # step1: 显示样本总条数
            print(f'Total Sample Num: {total_sample_num}')

            # step2: 显示采样的样本行索引
            print('Sample Index: ')
            print(sample_indexs)

            # step3: 显示采样的样本
            print('Samples: ')
            for s in samples:
                print(f'{s}\n')


# extract_from_coco('/Volumes/Elements/手势/gesture/annotations/dumix_test_face_merge_200622_bin.json', './', 'hand:1')
# extract_from_coco('/Volumes/Elements/手势/gesture/annotations/dumix_train_face_201117_tri.json', './', 'hand:1')
# print('sdf')


# Copyright (c) OpenMMLab. All rights reserved.
import functools
import pickle
import warnings
from collections import OrderedDict

import torch
import torch.distributed as dist
from antgo.framework.helper.runner import OptimizerHook, get_dist_info
from torch._utils import (_flatten_dense_tensors, _take_tensors,
                          _unflatten_dense_tensors)


def _allreduce_coalesced(tensors, world_size, bucket_size_mb=-1):
    if bucket_size_mb > 0:
        bucket_size_bytes = bucket_size_mb * 1024 * 1024
        buckets = _take_tensors(tensors, bucket_size_bytes)
    else:
        buckets = OrderedDict()
        for tensor in tensors:
            tp = tensor.type()
            if tp not in buckets:
                buckets[tp] = []
            buckets[tp].append(tensor)
        buckets = buckets.values()

    for bucket in buckets:
        flat_tensors = _flatten_dense_tensors(bucket)
        dist.all_reduce(flat_tensors)
        flat_tensors.div_(world_size)
        for tensor, synced in zip(
                bucket, _unflatten_dense_tensors(flat_tensors, bucket)):
            tensor.copy_(synced)


def allreduce_grads(params, coalesce=True, bucket_size_mb=-1):
    """Allreduce gradients.

    Args:
        params (list[torch.Parameters]): List of parameters of a model
        coalesce (bool, optional): Whether allreduce parameters as a whole.
            Defaults to True.
        bucket_size_mb (int, optional): Size of bucket, the unit is MB.
            Defaults to -1.
    """
    grads = [
        param.grad.data for param in params
        if param.requires_grad and param.grad is not None
    ]
    world_size = dist.get_world_size()
    if coalesce:
        _allreduce_coalesced(grads, world_size, bucket_size_mb)
    else:
        for tensor in grads:
            dist.all_reduce(tensor.div_(world_size))


class DistOptimizerHook(OptimizerHook):
    """Deprecated optimizer hook for distributed training."""

    def __init__(self, *args, **kwargs):
        warnings.warn('"DistOptimizerHook" is deprecated, please switch to'
                      '"mmcv.runner.OptimizerHook".')
        super().__init__(*args, **kwargs)


def reduce_mean(tensor):
    """"Obtain the mean of tensor on different GPUs."""
    if not (dist.is_available() and dist.is_initialized()):
        return tensor
    tensor = tensor.clone()
    dist.all_reduce(tensor.div_(dist.get_world_size()), op=dist.ReduceOp.SUM)
    return tensor


def obj2tensor(pyobj, device='cuda'):
    """Serialize picklable python object to tensor."""
    storage = torch.ByteStorage.from_buffer(pickle.dumps(pyobj))
    return torch.ByteTensor(storage).to(device=device)


def tensor2obj(tensor):
    """Deserialize tensor to picklable python object."""
    return pickle.loads(tensor.cpu().numpy().tobytes())


@functools.lru_cache()
def _get_global_gloo_group():
    """Return a process group based on gloo backend, containing all the ranks
    The result is cached."""
    if dist.get_backend() == 'nccl':
        return dist.new_group(backend='gloo')
    else:
        return dist.group.WORLD


def all_reduce_dict(py_dict, op='sum', group=None, to_float=True):
    """Apply all reduce function for python dict object.

    The code is modified from https://github.com/Megvii-
    BaseDetection/YOLOX/blob/main/yolox/utils/allreduce_norm.py.

    NOTE: make sure that py_dict in different ranks has the same keys and
    the values should be in the same shape. Currently only supports
    nccl backend.

    Args:
        py_dict (dict): Dict to be applied all reduce op.
        op (str): Operator, could be 'sum' or 'mean'. Default: 'sum'
        group (:obj:`torch.distributed.group`, optional): Distributed group,
            Default: None.
        to_float (bool): Whether to convert all values of dict to float.
            Default: True.

    Returns:
        OrderedDict: reduced python dict object.
    """
    warnings.warn(
        'group` is deprecated. Currently only supports NCCL backend.')
    _, world_size = get_dist_info()
    if world_size == 1:
        return py_dict

    # all reduce logic across different devices.
    py_key = list(py_dict.keys())
    if not isinstance(py_dict, OrderedDict):
        py_key_tensor = obj2tensor(py_key)
        dist.broadcast(py_key_tensor, src=0)
        py_key = tensor2obj(py_key_tensor)

    tensor_shapes = [py_dict[k].shape for k in py_key]
    tensor_numels = [py_dict[k].numel() for k in py_key]

    if to_float:
        warnings.warn('Note: the "to_float" is True, you need to '
                      'ensure that the behavior is reasonable.')
        flatten_tensor = torch.cat(
            [py_dict[k].flatten().float() for k in py_key])
    else:
        flatten_tensor = torch.cat([py_dict[k].flatten() for k in py_key])

    dist.all_reduce(flatten_tensor, op=dist.ReduceOp.SUM)
    if op == 'mean':
        flatten_tensor /= world_size

    split_tensors = [
        x.reshape(shape) for x, shape in zip(
            torch.split(flatten_tensor, tensor_numels), tensor_shapes)
    ]
    out_dict = {k: v for k, v in zip(py_key, split_tensors)}
    if isinstance(py_dict, OrderedDict):
        out_dict = OrderedDict(out_dict)
    return out_dict



from .utils import with_metaclass


class DefaultConfig:
    TARGET = 'antgo_op'
    BUILD_PATH = './.temp/'
    BUILD_IN_LOCAL_PATH = True
    SHOW_BUILDING_COMMAND = False
    MAX_BUILDING_WORKER_NUM = 8

    DEBUG = False
    USING_OPENMP = False
    USING_CBLAS = False
    HOST_NUM_THREADS = 0  # 0 : auto
    USING_HIGH_LEVEL_WARNINGS = False
    USING_OPTIMIZATION = True
    USING_ASYNC_EXEC = True
    GPU_BACKEND = 'cuda'

    CXX = 'g++'
    NVCC = 'nvcc'
    HIPCC = 'hipcc'
    CUDA_DIR = '/opt/cuda'
    HIP_DIR = '/opt/rocm/hip'


class Config:
    def __init__(self):
        for name in dir(DefaultConfig):
            if not name.startswith('_'):
                self.__dict__[name] = getattr(DefaultConfig, name)

    def __setattr__(self, name, value):
        data = self.__dict__.get(name, None)
        if data is None:
            raise AttributeError("Config has no attribute '{}'".format(name))
        target_type = type(data)
        value_type = type(value)
        if target_type is not value_type:
            raise TypeError('The type of config attribute `{}` is not consistent, target {} vs value {}.'.format(
                name, target_type, value_type))
        self.__dict__[name] = value


config = Config()


class TempConfig:
    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.old_config = dict()

    def __enter__(self):
        for k, v in self.kwargs.items():
            if not hasattr(config, k):
                raise AttributeError(
                    "'mobula.config' object has no attribute '{}'".format(k))
            self.old_config[k] = getattr(config, k)
            setattr(config, k, v)

    def __exit__(self, *dummy):
        for k, v in self.old_config.items():
            setattr(config, k, v)


Config.TempConfig = TempConfig



# Copyright (c) OpenMMLab. All rights reserved.
from functools import partial

import numpy as np
import torch
from six.moves import map, zip

from ..mask.structures import BitmapMasks, PolygonMasks


def multi_apply(func, *args, **kwargs):
    """Apply function to a list of arguments.

    Note:
        This function applies the ``func`` to multiple inputs and
        map the multiple outputs of the ``func`` into different
        list. Each list contains the same type of outputs corresponding
        to different inputs.

    Args:
        func (Function): A function that will be applied to a list of
            arguments

    Returns:
        tuple(list): A tuple containing multiple list, each list contains \
            a kind of returned results by the function
    """
    pfunc = partial(func, **kwargs) if kwargs else func
    map_results = map(pfunc, *args)
    return tuple(map(list, zip(*map_results)))


def unmap(data, count, inds, fill=0):
    """Unmap a subset of item (data) back to the original set of items (of size
    count)"""
    if data.dim() == 1:
        ret = data.new_full((count, ), fill)
        ret[inds.type(torch.bool)] = data
    else:
        new_size = (count, ) + data.size()[1:]
        ret = data.new_full(new_size, fill)
        ret[inds.type(torch.bool), :] = data
    return ret


def mask2ndarray(mask):
    """Convert Mask to ndarray..

    Args:
        mask (:obj:`BitmapMasks` or :obj:`PolygonMasks` or
        torch.Tensor or np.ndarray): The mask to be converted.

    Returns:
        np.ndarray: Ndarray mask of shape (n, h, w) that has been converted
    """
    if isinstance(mask, (BitmapMasks, PolygonMasks)):
        mask = mask.to_ndarray()
    elif isinstance(mask, torch.Tensor):
        mask = mask.detach().cpu().numpy()
    elif not isinstance(mask, np.ndarray):
        raise TypeError(f'Unsupported {type(mask)} data type')
    return mask


def flip_tensor(src_tensor, flip_direction):
    """flip tensor base on flip_direction.

    Args:
        src_tensor (Tensor): input feature map, shape (B, C, H, W).
        flip_direction (str): The flipping direction. Options are
          'horizontal', 'vertical', 'diagonal'.

    Returns:
        out_tensor (Tensor): Flipped tensor.
    """
    assert src_tensor.ndim == 4
    valid_directions = ['horizontal', 'vertical', 'diagonal']
    assert flip_direction in valid_directions
    if flip_direction == 'horizontal':
        out_tensor = torch.flip(src_tensor, [3])
    elif flip_direction == 'vertical':
        out_tensor = torch.flip(src_tensor, [2])
    else:
        out_tensor = torch.flip(src_tensor, [2, 3])
    return out_tensor


def select_single_mlvl(mlvl_tensors, batch_id, detach=True):
    """Extract a multi-scale single image tensor from a multi-scale batch
    tensor based on batch index.

    Note: The default value of detach is True, because the proposal gradient
    needs to be detached during the training of the two-stage model. E.g
    Cascade Mask R-CNN.

    Args:
        mlvl_tensors (list[Tensor]): Batch tensor for all scale levels,
           each is a 4D-tensor.
        batch_id (int): Batch index.
        detach (bool): Whether detach gradient. Default True.

    Returns:
        list[Tensor]: Multi-scale single image tensor.
    """
    assert isinstance(mlvl_tensors, (list, tuple))
    num_levels = len(mlvl_tensors)

    if detach:
        mlvl_tensor_list = [
            mlvl_tensors[i][batch_id].detach() for i in range(num_levels)
        ]
    else:
        mlvl_tensor_list = [
            mlvl_tensors[i][batch_id] for i in range(num_levels)
        ]
    return mlvl_tensor_list


def filter_scores_and_topk(scores, score_thr, topk, results=None):
    """Filter results using score threshold and topk candidates.

    Args:
        scores (Tensor): The scores, shape (num_bboxes, K).
        score_thr (float): The score filter threshold.
        topk (int): The number of topk candidates.
        results (dict or list or Tensor, Optional): The results to
           which the filtering rule is to be applied. The shape
           of each item is (num_bboxes, N).

    Returns:
        tuple: Filtered results

            - scores (Tensor): The scores after being filtered, \
                shape (num_bboxes_filtered, ).
            - labels (Tensor): The class labels, shape \
                (num_bboxes_filtered, ).
            - anchor_idxs (Tensor): The anchor indexes, shape \
                (num_bboxes_filtered, ).
            - filtered_results (dict or list or Tensor, Optional): \
                The filtered results. The shape of each item is \
                (num_bboxes_filtered, N).
    """
    valid_mask = scores > score_thr
    scores = scores[valid_mask]
    valid_idxs = torch.nonzero(valid_mask)

    num_topk = min(topk, valid_idxs.size(0))
    # torch.sort is actually faster than .topk (at least on GPUs)
    scores, idxs = scores.sort(descending=True)
    scores = scores[:num_topk]
    topk_idxs = valid_idxs[idxs[:num_topk]]
    keep_idxs, labels = topk_idxs.unbind(dim=1)

    filtered_results = None
    if results is not None:
        if isinstance(results, dict):
            filtered_results = {k: v[keep_idxs] for k, v in results.items()}
        elif isinstance(results, list):
            filtered_results = [result[keep_idxs] for result in results]
        elif isinstance(results, torch.Tensor):
            filtered_results = results[keep_idxs]
        else:
            raise NotImplementedError(f'Only supports dict or list or Tensor, '
                                      f'but get {type(results)}.')
    return scores, labels, keep_idxs, filtered_results


def center_of_mass(mask, esp=1e-6):
    """Calculate the centroid coordinates of the mask.

    Args:
        mask (Tensor): The mask to be calculated, shape (h, w).
        esp (float): Avoid dividing by zero. Default: 1e-6.

    Returns:
        tuple[Tensor]: the coordinates of the center point of the mask.

            - center_h (Tensor): the center point of the height.
            - center_w (Tensor): the center point of the width.
    """
    h, w = mask.shape
    grid_h = torch.arange(h, device=mask.device)[:, None]
    grid_w = torch.arange(w, device=mask.device)
    normalizer = mask.sum().float().clamp(min=esp)
    center_h = (mask * grid_h).sum() / normalizer
    center_w = (mask * grid_w).sum() / normalizer
    return center_h, center_w


def generate_coordinate(featmap_sizes, device='cuda'):
    """Generate the coordinate.

    Args:
        featmap_sizes (tuple): The feature to be calculated,
            of shape (N, C, W, H).
        device (str): The device where the feature will be put on.
    Returns:
        coord_feat (Tensor): The coordinate feature, of shape (N, 2, W, H).
    """

    x_range = torch.linspace(-1, 1, featmap_sizes[-1], device=device)
    y_range = torch.linspace(-1, 1, featmap_sizes[-2], device=device)
    y, x = torch.meshgrid(y_range, x_range)
    y = y.expand([featmap_sizes[0], 1, -1, -1])
    x = x.expand([featmap_sizes[0], 1, -1, -1])
    coord_feat = torch.cat([x, y], 1)

    return coord_feat



# -*- coding: UTF-8 -*-
# @Time    : 2022/9/13 16:10
# @File    : yolov7.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import os

import numpy as np
import torch
import torch.nn as nn
import yaml

from antgo.pipeline.models.utils.utils import *
from antgo.pipeline.models.utils.preprocess import *
from .util import *
from ..util import *


class Model(nn.Module):
  def __init__(self, model_name='yolov7', device='cpu'):
    super().__init__()
    self.device = device
    model_path = torch_model_download(model_name, file_type='.pth')

    pure_model_name = model_name.split('.')[0]
    model_cfg = \
      os.path.realpath(__file__).split('/')[:-3] + ['cfg', 'detector', f'{pure_model_name}.yaml']
    model_cfg_file = '/'.join(model_cfg)
    assert(os.path.exists(model_cfg_file))
    with open(model_cfg_file) as f:
      self.yaml = yaml.load(f, Loader=yaml.SafeLoader)  # model dict

    # 解析模型，并加载权重
    self.model, self.save = yolo_parse_model(self.yaml, ch=[3])
    self.fuse()
    self.to(self.device)
    self.load_state_dict(torch.load(model_path, map_location=self.device))

    # Build strides, anchors
    m = self.model[-1]  # Detect()
    m.stride = torch.tensor([8,16,32])
    m.anchors /= m.stride.view(-1, 1, 1)
    self.stride = m.stride

  def forward(self, x):
    y = []
    for m in self.model:
      if m.f != -1:  # if not from previous layer
        x = y[m.f] if isinstance(m.f, int) else [x if j == -1 else y[j] for j in m.f]  # from earlier layers

      x = m(x)  # run

      y.append(x if m.i in self.save else None)  # save output
    return x

  def fuse(self):
    for m in self.model.modules():
      if isinstance(m, RepConv):
        m.fuse_repvgg_block()
      elif isinstance(m, RepConv_OREPA):
        m.switch_to_deploy()
      elif type(m) is Conv and hasattr(m, 'bn'):
        m.conv = fuse_conv_and_bn(m.conv, m.bn)  # update conv
        delattr(m, 'bn')  # remove batchnorm
        m.forward = m.fuseforward  # update forward
    return self


class Yolov7(object):
  def __init__(self, model_name='yolov7', device='cpu', conf_thres=0.25, iou_thres=0.45, dataset='coco'):
    self.device = select_device(device)
    self.model = Model(f'{model_name}.{dataset.lower()}', self.device)
    self.conf_thres = conf_thres
    self.iou_thres = iou_thres

    dataset_cfg = \
      os.path.realpath(__file__).split('/')[:-3] + ['cfg', 'dataset', f'{dataset.lower()}.yaml']
    dataset_cfg_file = '/'.join(dataset_cfg)
    with open(dataset_cfg_file) as f:
      self.yaml = yaml.load(f, Loader=yaml.SafeLoader)  # model dict
    self.names = self.yaml['names']

  def __call__(self, *args, **kwargs):
    img0 = args[0]
    img0_shape = img0.shape
    # Padded resize
    img = paddedResize(img0, (640, 640), stride=32)[0]
    img_shape = img.shape

    # Convert
    img = img[:, :, ::-1].transpose(2, 0, 1)  # BGR to RGB, to 3x416x416
    img = np.ascontiguousarray(img)

    img = torch.from_numpy(img).to(self.device)
    img = img.float()
    img /= 255.0  # 0 - 255 to 0.0 - 1.0
    if img.ndimension() == 3:
      img = img.unsqueeze(0)

    pred = self.model(img)[0]

    # Apply NMS
    pred = non_max_suppression(pred, self.conf_thres, self.iou_thres, None, agnostic=True)

    if len(pred) == 0:
      return np.empty((0,5), dtype=np.float32), []

    det = pred[0]
    # Rescale boxes from img_size to im0 size
    det[:, :4] = scale_coords(img_shape[:2], det[:, :4], img0_shape[:2]).round()

    det_names = []
    for *xyxy, conf, cls in reversed(det):
      cls_name = self.names[(int)(cls)]
      det_names.append(cls_name)

    det = det.detach().cpu().numpy()
    return det, det_names



import os
import logging
import json

def filter_by_tags(src_file, tgt_file, tags, no_tags):
    # use browser filter
    # src check
    if not os.path.exists(src_file):
        logging.error(f"{src_file} not existed")
        return
    if not src_file.endswith('.json'):
        logging.error(f"{src_file} must be json file")
        return
    
    if tags is None and no_tags is None:
        logging.error("Must set tags or no-tags")
        return
    
    tags_text = None
    no_tags_text = None
    if tags is not None:
        tags = set(tags.split(','))
        tags_text = '_'.join(tags)
    if no_tags is not None:
        no_tags = set(no_tags.split(','))
        no_tags_text = '_'.join(no_tags)
    
    with open(src_file, 'r') as fp:
        sample_list = json.load(fp)
    
    with open(tgt_file, 'r') as fp:
        filter_condition = json.load(fp)
    
    remain_sample_list = []
    for _, v in filter_condition.items():   
        tag_set = set(v[0]['tag'])
        if tags is not None:
            if not(tag_set == tags):
                continue
        if no_tags is not None:
            if tag_set == no_tags:
                continue
            
        sample = sample_list[int(v[0]['id'])]
        remain_sample_list.append(sample)
    
    src_folder = os.path.dirname(src_file)
    src_name = src_file.split('/')[-1].split('.')[0]
    
    prefix = f'{src_name}_filter'
    if tags is not None:
        prefix = f'{prefix}_tags_{tags_text}'
    if no_tags is not None:
        prefix = f'{prefix}_no_tags_{no_tags_text}'
    
    with open(os.path.join(src_folder, f'{prefix}.json'), 'w') as fp:
        json.dump(remain_sample_list,fp)







# -*- coding: UTF-8 -*-
# @Time    : 2022/9/12 21:53
# @File    : data_processing.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function

import random
from antgo.pipeline.functional.mixins.dag import register_dag
from typing import Iterable

from antgo.pipeline.functional.entity import Entity


class DataProcessingMixin:
    """
    Mixin for processing data.
    """
    @register_dag
    def select_from(self, other):
        """
        Select data from dc with list(self).

        Examples:

        >>> from antgo.pipeline import DataCollection
        >>> dc1 = DataCollection([0.8, 0.9, 8.1, 9.2])
        >>> dc2 = DataCollection([[1, 2, 0], [2, 3, 0]])

        >>> dc3 = dc2.select_from(dc1)
        >>> list(dc3)
        [[0.9, 8.1, 0.8], [8.1, 9.2, 0.8]]
        """
        self.parent_ids.append(other.id)
        other.notify_consumed(self.id)

        def inner(x):
            if isinstance(x, Iterable):
                return [other[i] for i in x]
            return other[x]

        result = map(inner, self._iterable)
        return self._factory(result)

    @register_dag
    def zip(self, *others) -> 'DataCollection':
        """
        Combine two data collections.

        Args:
            *others (DataCollection): other data collections;

        Returns:
            DataCollection: data collection with zipped values;

        Examples:

        >>> from antgo.pipeline import DataCollection
        >>> dc1 = DataCollection([1,2,3,4])
        >>> dc2 = DataCollection([1,2,3,4]).map(lambda x: x+1)
        >>> dc3 = dc1.zip(dc2)
        >>> list(dc3)
        [(1, 2), (2, 3), (3, 4), (4, 5)]
        """
        self.parent_ids.extend([other.id for other in others])

        for x in others:
            x.notify_consumed(self.id)

        return self._factory(zip(self, *others))

    @register_dag
    def head(self, n: int = 5):
        """
        Get the first n lines of a DataCollection.

        Args:
            n (`int`):
                The number of lines to print. Default value is 5.

        Examples:

        >>> from antgo.pipeline import DataCollection
        >>> DataCollection.range(10).head(3).to_list()
        [0, 1, 2]
        """
        def inner():
            for i, x in enumerate(self._iterable):
                if i >= n:
                    break
                yield x

        return self._factory(inner())

    @register_dag
    def sample(self, ratio=1.0) -> 'DataCollection':
        """
        Sample the data collection.

        Args:
            ratio (float): sample ratio;

        Returns:
            DataCollection: sampled data collection;

        Examples:

        >>> from antgo.pipeline import DataCollection
        >>> dc = DataCollection(range(10000))
        >>> result = dc.sample(0.1)
        >>> ratio = len(result.to_list()) / 10000.
        >>> 0.09 < ratio < 0.11
        True
        """
        return self._factory(filter(lambda _: random.random() < ratio, self))

    @register_dag
    def batch(self, size, drop_tail=False, raw=True):
        """
        Create small batches from data collections.

        Args:
            size (`int`):
                Window size;
            drop_tail (`bool`):
                Drop tailing windows that not full, defaults to False;
            raw (`bool`):
                Whether to return raw data instead of DataCollection, defaults to True

        Returns:
            DataCollection of batched windows or batch raw data

        Examples:

        >>> from antgo.pipeline import DataCollection
        >>> dc = DataCollection(range(10))
        >>> [list(batch) for batch in dc.batch(2, raw=False)]
        [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]]

        >>> dc = DataCollection(range(10))
        >>> dc.batch(3)
        [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]

        >>> dc = DataCollection(range(10))
        >>> dc.batch(3, drop_tail=True)
        [[0, 1, 2], [3, 4, 5], [6, 7, 8]]

        >>> from antgo.pipeline import Entity
        >>> dc = DataCollection([Entity(a=a, b=b) for a,b in zip(['abc', 'vdfvcd', 'cdsc'], [1,2,3])])
        >>> dc.batch(2)
        [<Entity dict_keys(['a', 'b'])>, <Entity dict_keys(['a', 'b'])>]
        """
        def inner():
            buff = []
            count = 0
            for ele in self._iterable:
                if isinstance(ele, Entity):
                    if count == 0:
                        buff = ele
                        for key in ele.__dict__.keys():
                            buff.__dict__[key] = [buff.__dict__[key]]
                        count = 1
                        continue
                    for key in ele.__dict__.keys():
                        buff.__dict__[key].append(ele.__dict__[key])
                else:
                    buff.append(ele)
                count += 1

                if count == size:
                    if raw:
                        yield buff
                    else:
                        yield buff if isinstance(buff, list) else [buff]
                    buff = []
                    count = 0
            if not drop_tail and count > 0:
                if raw:
                    yield buff
                else:
                    yield buff if isinstance(buff, list) else [buff]

        return self._factory(inner())

    @register_dag
    def rolling(self, size: int, drop_head=True, drop_tail=True):
        """
        Create rolling windows from data collections.

        Args:
            size (`int`):
                Wndow size.
            drop_head (`bool`):
                Drop headding windows that not full.
            drop_tail (`bool`):
                Drop tailing windows that not full.

        Returns:
            DataCollection: data collection of rolling windows;

        Examples:

        >>> from antgo.pipeline import DataCollection
        >>> dc = DataCollection(range(5))
        >>> [list(batch) for batch in dc.rolling(3)]
        [[0, 1, 2], [1, 2, 3], [2, 3, 4]]

        >>> dc = DataCollection(range(5))
        >>> [list(batch) for batch in dc.rolling(3, drop_head=False)]
        [[0], [0, 1], [0, 1, 2], [1, 2, 3], [2, 3, 4]]

        >>> dc = DataCollection(range(5))
        >>> [list(batch) for batch in dc.rolling(3, drop_tail=False)]
        [[0, 1, 2], [1, 2, 3], [2, 3, 4], [3, 4], [4]]
        """
        def inner():
            buff = []
            for ele in self._iterable:
                buff.append(ele)
                if not drop_head or len(buff) == size:
                    yield buff.copy()
                if len(buff) == size:
                    buff = buff[1:]
            while not drop_tail and len(buff) > 0:
                yield buff
                buff = buff[1:]

        return self._factory(inner())

    @register_dag
    def flatten(self) -> 'DataCollection':
        """
        Flatten nested data collections.

        Returns:
            DataCollection: flattened data collection;

        Examples:

        >>> from antgo.pipeline import DataCollection
        >>> dc = DataCollection(range(10))
        >>> nested_dc = dc.batch(2)
        >>> nested_dc.flatten().to_list()
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        """
        def inner():
            for ele in self._iterable:
                if isinstance(ele, Iterable):
                    for nested_ele in iter(ele):
                        yield nested_ele
                else:
                    yield ele

        return self._factory(inner())

    @register_dag
    def shuffle(self) -> 'DataCollection':
        """
        Shuffle an unstreamed data collection in place.

        Returns:
            DataCollection: shuffled data collection;

        Examples:

        1. Shuffle:

        >>> from antgo.pipeline import DataCollection
        >>> dc = DataCollection([0, 1, 2, 3, 4])
        >>> a = dc.shuffle()
        >>> tuple(a) == tuple(range(5))
        False

        2. streamed data collection is not supported:

        >>> dc = DataCollection([0, 1, 2, 3, 4]).stream()
        >>> _ = dc.shuffle()
        Traceback (most recent call last):
        TypeError: shuffle is not supported for streamed data collection.
        """
        if self.is_stream:
            raise TypeError('shuffle is not supported for streamed data collection.')
        iterable = random.sample(self._iterable, len(self._iterable))
        return self._factory(iterable)



import bisect
import collections
import copy
import math
from collections import defaultdict
from xml.etree.ElementTree import iselement

import numpy as np
from antgo.framework.helper.utils import build_from_cfg, print_log
import torch
from torch.utils.data.dataset import ConcatDataset as _ConcatDataset
from .builder import DATASETS, PIPELINES


@DATASETS.register_module()
class ConcatDataset(_ConcatDataset):
    """A wrapper of concatenated dataset.

    Same as :obj:`torch.utils.data.dataset.ConcatDataset`, but
    concat the group flag for image aspect ratio.

    Args:
        datasets (list[:obj:`Dataset`]): A list of datasets.
    """

    def __init__(self, datasets):
        super(ConcatDataset, self).__init__(datasets)
        self.CLASSES = getattr(datasets[0], 'CLASSES', None)
        self.PALETTE = getattr(datasets[0], 'PALETTE', None)

        self.flag = []
        for index, dataset in enumerate(datasets):
            self.flag.append(np.ones(len(dataset), dtype=np.int64) * index)
        self.flag = np.concatenate(self.flag)

    def __getitem__(self, idx):
        if type(idx) == list:
            # 聚合到每个数据集
            d_i_map = {}
            for i in idx:
                if i < 0:
                    if -i > len(self):
                        raise ValueError("absolute value of index should not exceed dataset length")
                    i = len(self) + i

                dataset_idx = bisect.bisect_right(self.cumulative_sizes, i)
                if dataset_idx not in d_i_map:
                    d_i_map[dataset_idx] = []

                if dataset_idx == 0:
                    sample_idx = i
                else:
                    sample_idx = i - self.cumulative_sizes[dataset_idx - 1]
                                    
                d_i_map[dataset_idx].append(sample_idx)

            sample_list = []
            for dataset_idx, sample_idxs in d_i_map.items():
                sample_list.extend(self.datasets[dataset_idx][sample_idxs])
            
            return sample_list
        else:
            # 
            if idx < 0:
                if -idx > len(self):
                    raise ValueError("absolute value of index should not exceed dataset length")
                idx = len(self) + idx
            dataset_idx = bisect.bisect_right(self.cumulative_sizes, idx)
            if dataset_idx == 0:
                sample_idx = idx
            else:
                sample_idx = idx - self.cumulative_sizes[dataset_idx - 1]
            return self.datasets[dataset_idx][sample_idx]

    def get_cat_ids(self, idx):
        """Get category ids of concatenated dataset by index.

        Args:
            idx (int): Index of data.

        Returns:
            list[int]: All categories in the image of specified index.
        """

        if idx < 0:
            if -idx > len(self):
                raise ValueError(
                    'absolute value of index should not exceed dataset length')
            idx = len(self) + idx
        dataset_idx = bisect.bisect_right(self.cumulative_sizes, idx)
        if dataset_idx == 0:
            sample_idx = idx
        else:
            sample_idx = idx - self.cumulative_sizes[dataset_idx - 1]
        return self.datasets[dataset_idx].get_cat_ids(sample_idx)

    def get_ann_info(self, idx):
        """Get annotation of concatenated dataset by index.

        Args:
            idx (int): Index of data.

        Returns:
            dict: Annotation info of specified index.
        """

        if idx < 0:
            if -idx > len(self):
                raise ValueError(
                    'absolute value of index should not exceed dataset length')
            idx = len(self) + idx
        dataset_idx = bisect.bisect_right(self.cumulative_sizes, idx)
        if dataset_idx == 0:
            sample_idx = idx
        else:
            sample_idx = idx - self.cumulative_sizes[dataset_idx - 1]
        return self.datasets[dataset_idx].get_ann_info(sample_idx)

    def worker_init_fn(self, *args, **kwargs):
        for dataset in self.datasets:
            if getattr(dataset, 'worker_init_fn', None):
                dataset.worker_init_fn(*args, **kwargs)
    @property
    def is_kv(self):
        return getattr(self.datasets[0], 'is_kv', False)


@DATASETS.register_module()
class IterConcatDataset(torch.utils.data.ChainDataset):
    def __init__(self, datasets, samples_per_gpu=None) -> None:
        super().__init__(datasets)
        self.samples_per_gpu = samples_per_gpu

    def __iter__(self):
        dataset_iters = [iter(d) for d in self.datasets]

        while True:
            data_list = []
            is_exit = False
            for dataset, batch_size in zip(dataset_iters, self.samples_per_gpu):
                for _ in range(batch_size):
                    try:
                        data_list.append(next(dataset))
                        
                    except StopIteration:
                        is_exit = True
                
                if is_exit:
                    break
            if is_exit:
                break
            
            for data in data_list:
                yield data

    def __len__(self):
        total = 0
        for d in self.datasets:
            d_len = len(d)
            if total < d_len:
                total = d_len

        return total



@DATASETS.register_module()
class RepeatDataset:
    """A wrapper of repeated dataset.

    The length of repeated dataset will be `times` larger than the original
    dataset. This is useful when the data loading time is long but the dataset
    is small. Using RepeatDataset can reduce the data loading time between
    epochs.

    Args:
        dataset (:obj:`Dataset`): The dataset to be repeated.
        times (int): Repeat times.
    """

    def __init__(self, dataset, times):
        self.dataset = dataset
        self.times = times
        self.CLASSES = getattr(dataset, 'CLASSES', None)
        self.PALETTE = getattr(dataset, 'PALETTE', None)
        if hasattr(self.dataset, 'flag'):
            self.flag = np.tile(self.dataset.flag, times)

        self._ori_len = len(self.dataset)

    def __getitem__(self, idx):
        return self.dataset[idx % self._ori_len]

    def get_cat_ids(self, idx):
        """Get category ids of repeat dataset by index.

        Args:
            idx (int): Index of data.

        Returns:
            list[int]: All categories in the image of specified index.
        """

        return self.dataset.get_cat_ids(idx % self._ori_len)

    def get_ann_info(self, idx):
        """Get annotation of repeat dataset by index.

        Args:
            idx (int): Index of data.

        Returns:
            dict: Annotation info of specified index.
        """

        return self.dataset.get_ann_info(idx % self._ori_len)

    def __len__(self):
        """Length after repetition."""
        return self.times * self._ori_len




# -*- coding: UTF-8 -*-
# @Time    : 2022/9/12 14:19
# @File    : util.py
# @Author  : jian<jian@mltalker.com>
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from antgo.pipeline.engine import *
import cv2
import os
import base64
import numpy as np

@register 
class stack(object):
    def __init__(self, axis=0, transpose_dims=None) -> None:
        self.axis = axis
        self.transpose_dims = transpose_dims
    
    def __call__(self, x):
        x = np.stack(x, self.axis)
        if self.transpose_dims is not None:
            x = np.transpose(x, self.transpose_dims)
        return x


from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import torch
import torch.nn as nn
import torch.nn.functional as F
from ...builder import DISTILL_LOSSES


@DISTILL_LOSSES.register_module()
class IRG(nn.Module):
	'''
	Knowledge Distillation via Instance Relationship Graph
	http://openaccess.thecvf.com/content_CVPR_2019/papers/
	Liu_Knowledge_Distillation_via_Instance_Relationship_Graph_CVPR_2019_paper.pdf
	The official code is written by Caffe
	https://github.com/yufanLIU/IRG
	'''
	def __init__(self, w_irg_vert, w_irg_edge, w_irg_tran):
		super(IRG, self).__init__()

		self.w_irg_vert = w_irg_vert
		self.w_irg_edge = w_irg_edge
		self.w_irg_tran = w_irg_tran

	def forward(self, irg_s, irg_t):
		fm_s1, fm_s2, feat_s, out_s = irg_s
		fm_t1, fm_t2, feat_t, out_t = irg_t

		loss_irg_vert = F.mse_loss(out_s, out_t)

		irg_edge_feat_s = self.euclidean_dist_feat(feat_s, squared=True)
		irg_edge_feat_t = self.euclidean_dist_feat(feat_t, squared=True)
		irg_edge_fm_s1  = self.euclidean_dist_fm(fm_s1, squared=True)
		irg_edge_fm_t1  = self.euclidean_dist_fm(fm_t1, squared=True)
		irg_edge_fm_s2  = self.euclidean_dist_fm(fm_s2, squared=True)
		irg_edge_fm_t2  = self.euclidean_dist_fm(fm_t2, squared=True)
		loss_irg_edge = (F.mse_loss(irg_edge_feat_s, irg_edge_feat_t) +
						 F.mse_loss(irg_edge_fm_s1,  irg_edge_fm_t1 ) +
						 F.mse_loss(irg_edge_fm_s2,  irg_edge_fm_t2 )) / 3.0

		irg_tran_s = self.euclidean_dist_fms(fm_s1, fm_s2, squared=True)
		irg_tran_t = self.euclidean_dist_fms(fm_t1, fm_t2, squared=True)
		loss_irg_tran = F.mse_loss(irg_tran_s, irg_tran_t)

		# print(self.w_irg_vert * loss_irg_vert)
		# print(self.w_irg_edge * loss_irg_edge)
		# print(self.w_irg_tran * loss_irg_tran)
		# print()

		loss = (self.w_irg_vert * loss_irg_vert +
				self.w_irg_edge * loss_irg_edge +
				self.w_irg_tran * loss_irg_tran)

		return loss

	def euclidean_dist_fms(self, fm1, fm2, squared=False, eps=1e-12):
		'''
		Calculating the IRG Transformation, where fm1 precedes fm2 in the network.
		'''
		if fm1.size(2) > fm2.size(2):
			fm1 = F.adaptive_avg_pool2d(fm1, (fm2.size(2), fm2.size(3)))
		if fm1.size(1) < fm2.size(1):
			fm2 = (fm2[:,0::2,:,:] + fm2[:,1::2,:,:]) / 2.0

		fm1 = fm1.view(fm1.size(0), -1)
		fm2 = fm2.view(fm2.size(0), -1)
		fms_dist = torch.sum(torch.pow(fm1-fm2, 2), dim=-1).clamp(min=eps)

		if not squared:
			fms_dist = fms_dist.sqrt()

		fms_dist = fms_dist / fms_dist.max()

		return fms_dist

	def euclidean_dist_fm(self, fm, squared=False, eps=1e-12): 
		'''
		Calculating the IRG edge of feature map. 
		'''
		fm = fm.view(fm.size(0), -1)
		fm_square = fm.pow(2).sum(dim=1)
		fm_prod   = torch.mm(fm, fm.t())
		fm_dist   = (fm_square.unsqueeze(0) + fm_square.unsqueeze(1) - 2 * fm_prod).clamp(min=eps)

		if not squared:
			fm_dist = fm_dist.sqrt()

		fm_dist = fm_dist.clone()
		fm_dist[range(len(fm)), range(len(fm))] = 0
		fm_dist = fm_dist / fm_dist.max()

		return fm_dist

	def euclidean_dist_feat(self, feat, squared=False, eps=1e-12):
		'''
		Calculating the IRG edge of feat.
		'''
		feat_square = feat.pow(2).sum(dim=1)
		feat_prod   = torch.mm(feat, feat.t())
		feat_dist   = (feat_square.unsqueeze(0) + feat_square.unsqueeze(1) - 2 * feat_prod).clamp(min=eps)

		if not squared:
			feat_dist = feat_dist.sqrt()

		feat_dist = feat_dist.clone()
		feat_dist[range(len(feat)), range(len(feat))] = 0
		feat_dist = feat_dist / feat_dist.max()

		return feat_dist